Package: bsub
Type: Package
Title: Submitter and Monitor of the 'LSF Cluster'
Version: 2.0.7
Date: 2026-04-08
Authors@R: person("Zuguang", "Gu", email = "guzuguang@suat-sz.edu.cn", role = c("aut", "cre"),
                  comment = c('ORCID'="0000-0002-7395-8709"))
Depends: R (>= 4.0.0)
Imports: GlobalOptions (>= 0.1.1), GetoptLong (>= 0.1.8), digest,
        utils, stats, clisymbols, crayon, grDevices, graphics,
        codetools, ssh, igraph
Suggests: knitr, DT (>= 0.13), shiny (>= 1.0.0), shinyjs, graph,
        ggplot2, rmarkdown, markdown, DiagrammeR
biocViews: Software, Infrastructure
Description: It submits R code/R scripts/shell commands to 'LSF cluster' 
  (<https://en.wikipedia.org/wiki/Platform_LSF>, the 'bsub' system) without 
  leaving R. There is also an interactive 'shiny' application for monitoring job status.
VignetteBuilder: knitr
URL: https://github.com/jokergoo/bsub
License: MIT + file LICENSE
NeedsCompilation: no
SystemRequirements: Platform LSF, bsub
Encoding: UTF-8
RoxygenNote: 7.3.1
Packaged: 2026-04-08 04:52:52 UTC; guz
Author: Zuguang Gu [aut, cre] (ORCID: <https://orcid.org/0000-0002-7395-8709>)
Maintainer: Zuguang Gu <guzuguang@suat-sz.edu.cn>
Repository: CRAN
Date/Publication: 2026-04-08 09:40:14 UTC
Built: R 4.6.0; ; 2026-04-26 11:56:45 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,bconf)
S3method(print,bjobs)
export(bconf)
export(bjobs)
export(bjobs_barplot)
export(bjobs_done)
export(bjobs_exit)
export(bjobs_pending)
export(bjobs_raw)
export(bjobs_reset_timestamp)
export(bjobs_running)
export(bjobs_timeline)
export(bkill)
export(brecent)
export(bsub_chunk)
export(bsub_cmd)
export(bsub_opt)
export(bsub_script)
export(config_bsub)
export(formatFileSize)
export(job_dependency_all)
export(job_dependency_diagram)
export(job_dependency_dot)
export(job_dependency_igraph)
export(job_log)
export(job_rerun)
export(job_status_by_id)
export(job_status_by_name)
export(list_dump_files)
export(list_temp_files)
export(monitor)
export(pipeline_rerun)
export(random_job)
export(remove_temp_files)
export(retrieve_var)
export(run_cmd)
export(ssh_connect)
export(ssh_disconnect)
import(GetoptLong)
import(GlobalOptions)
import(clisymbols)
import(crayon)
import(igraph)
importFrom(codetools,findGlobals)
importFrom(grDevices,col2rgb)
importFrom(grDevices,rgb)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,legend)
importFrom(graphics,par)
importFrom(graphics,segments)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(ssh,ssh_session_info)
importFrom(stats,runif)
importFrom(utils,find)
importFrom(utils,getFromNamespace)
importFrom(utils,read.csv)
importFrom(utils,sessionInfo)

bsub
====

python wrapper to submit jobs to bsub (and later qsub)

Authors
------
@brentp, @brwnj


Example
-------

```python

>>> sub = bsub("some_job", R="rusage[mem=1]", verbose=True)

# submit a job via call'ing the sub object with the command to run.
# the return value is the numeric job id.
>>> print sub("date").isdigit()
True

# 2nd argument can be a shell script, in which case
# the call() is empty.
#>>> bsub("somejob", "run.sh", verbose=True)()

# dependencies:
>>> job_id = bsub("sleeper", verbose=True)("sleep 2")
>>> bsub.poll(job_id)
True

```

or use the command-line to poll for running jobs:


```Shell
python -m bsub 12345 12346 12347
```

will block until those 3 jobs finish.

bsub_opt$output_dir = "/desktop-home/guz/project/development/bsub/inst/bsub_test"
bsub_opt$debug = TRUE

# Send R code
bsub_chunk(name = "example_01", 
{
	Sys.sleep(10)
})

bsub_chunk(name = "example_02", R_version = "3.5.1",
{
	Sys.sleep(5)
})

bsub_chunk(name = "example_03", packages = c("circlize", "ComplexHeatmap"),
{
	Sys.sleep(5)
})


foo = 1
bsub_chunk(name = "example_04", variables = "foo",
{ 
	bar = foo
	Sys.sleep(5)
})

bsub_chunk(name = "example_05", save_var = TRUE,
{
	Sys.sleep(10)
	1+1
})
retrieve_var("example_05")

bsub_chunk(name = "load_cola",
{ 
	library(cola)
})


bsub_cmd("ls", name = "ls")

suppressPackageStartupMessages(library(DT))
suppressPackageStartupMessages(library(shiny))
suppressPackageStartupMessages(library(GetoptLong))
suppressPackageStartupMessages(library(DiagrammeR))
suppressPackageStartupMessages(library(igraph))
suppressPackageStartupMessages(library(shinyjs))
suppressPackageStartupMessages(library(bsub))

STATUS_COL = bsub:::STATUS_COL

ui = fluidPage(
    shinyjs::useShinyjs(),

    tags$style(HTML(
        paste(readLines("style.css"), collapse = "\n")
    )),
    tags$script(HTML(
        paste(readLines("app.js"), collapse = "\n")
    )),

    titlePanel("LSF Job Monitor"),

    htmlOutput("info"),
    actionButton("reload", "Manually reload"),

    # tags$script(HTML("count_down('countdown');")),
    
    hr(style = "border-top: 1px solid black;"),
    checkboxGroupInput("status_select", label = "Select job status: ", choices = structure(names(STATUS_COL), names = names(STATUS_COL)), selected = names(STATUS_COL), inline = TRUE),
    div(id = "table_loading", class = "loader"),
    DT::dataTableOutput("mytable"),
    HTML('<p><button id="kill" type="button" class="btn btn-default">Kill selected jobs</button></p>'),

    hr(style = "border-top: 1px solid black;"),
    p("built with ", a(qq("the bsub package v@{installed.packages()['bsub', 'Version']}"), href='https://github.com/jokergoo/bsub', target='_blank'))
)

ENV = new.env()
ENV$job_log = list()
ENV$timestamp = Sys.time()

server <- function(input, output, session) {
    
    # autoInvalidate <- reactiveTimer(1000 * 60*5)  # every 5 min
    
    # observe({
    #     t = autoInvalidate()

    #     if(as.double(t - ENV$timestamp, units = "secs") > 10) {
    #         message(qq("[@{format(Sys.time())}] Reloading the app."))
    #         ENV$timestamp = t
    #         session$reload() 
    #     }
    # })
        
    job_summary_table = reactive({
        
        # autoInvalidate()
        
        showNotification("Fetching job summary table...", duration = 1, type = "message")
        message(qq("[@{format(Sys.time())}] Fetching job summary table"))
        job_tb = bjobs(print = FALSE, status = "all")
        if(!is.null(job_tb)) {
            job_tb[order(job_tb$JOBID, decreasing = TRUE), , drop = FALSE]
        } else {
            job_tb
        }
    })

    observeEvent(input$reload, {
        ENV$timestamp = Sys.time()
        message(qq("[@{format(Sys.time())}] Reloading the app."))
        session$reload()  
    })

	output$mytable = DT::renderDataTable({
	    
	    # autoInvalidate()
	    
		job_tb = job_summary_table()
        if(is.null(job_tb)) {
            return(datatable(data.frame(numeric(0))))
        }
		nr = nrow(job_tb)

        job_dep = job_dependency_all(job_tb)
        
		showNotification("Formatting job summary table...", duration = 5, type = "message")
		job_tb2 = job_tb[, c("JOBID", "STAT", "JOB_NAME", "QUEUE", "SUBMIT_TIME", "TIME_PASSED", "RUNTIMELIMIT", "NREQ_SLOT", "MEM", "MAX_MEM", "REQ_MEM")]

		job_tb2$STAT = factor(job_tb2$STAT)
        units(job_tb2$TIME_PASSED) = "secs"; job_tb2$TIME_PASSED = as.integer(as.numeric(job_tb2$TIME_PASSED))
        job_tb2$RUNTIMELIMIT = job_tb2$RUNTIMELIMIT*60
        job_tb2$MAX_MEM = bsub:::convert_to_byte(job_tb2$MAX_MEM)
        job_tb2$MEM = bsub:::convert_to_byte(job_tb2$MEM)
        job_tb2$REQ_MEM = job_tb2$REQ_MEM*1024*1024*1024

        l = nchar(job_tb2$JOB_NAME) > 50
        if(any(l)) {
            foo = substr(job_tb2$JOB_NAME[l], 1, 48)
            foo = paste(foo, "..", sep = "")
            job_tb2$JOB_NAME[l] = foo
        }
        
        job_tb2$dep = ""

        ## add the dependency table
        if(any(job_tb2$JOBID %in% names(job_dep$id2name))) {
            for(i in which(job_tb2$JOBID %in% job_dep$dep_mat)) {
                job_tb2$dep[i] = "Dep"
            }
        }

        col_index_add = 0
        if(all(job_tb2$dep == "")) {
            job_tb2 = cbind(job_tb2[, 1, drop = FALSE], job_tb2[, -c(1, length(job_tb2)), drop = FALSE])
            colnames(job_tb2) = c("Job ID", "Status", "Job name", "Queue", "Submit time", "Passed time", "Req time", "Slots", "Mem", "Max mem", "Req mem")
        } else {
            col_index_add = 1
            job_tb2 = cbind(job_tb2[, c(1, length(job_tb2)), drop = FALSE], job_tb2[, -c(1, length(job_tb2)), drop = FALSE])
            colnames(job_tb2) = c("Job ID", "", "Status", "Job name", "Queue", "Submit time", "Passed time", "Req time", "Slots", "Mem", "Max mem", "Req mem")
		}
        job_tb2[, "Submit time"] = as.character(job_tb2[, "Submit time"])

        job_tb2 = job_tb2[job_tb2$Status %in% input$status_select, , drop = FALSE]

        if(nrow(job_tb2) == 0) {
            return(datatable(job_tb2, escape = FALSE, rownames = FALSE))
        }
        dt = datatable(job_tb2, escape = FALSE, rownames = FALSE, selection = 'none',
            options = list(
                pageLength = 25
            ), callback = JS("
table.on( 'draw', function () {
    $('#table_loading').hide();
    dt_bind_events();
});")
        )

        dt %>% 
            formatDate(columns = "Submit time", method = "toLocaleString") %>%
            bsub:::formatTimeDiff(columns = c("Passed time", "Req time")) %>%
            bsub:::formatNumber(columns = c("Slots")) %>%
            bsub:::formatFileSize(columns = c("Mem", "Max mem", "Req mem"))

	})
	
	output$info = renderText({
	    
	    # autoInvalidate()
	    
	    job_tb = job_summary_table()
	    tb = table(job_tb$STAT)
	    txt = paste0(qq("User <b>@{bsub_opt$user}</b> has "), paste(qq("<span class='@{names(tb)}' onclick='list_by_status(\"@{names(tb)}\");false;' style='cursor:pointer;'>@{tb} @{names(tb)} job@{ifelse(tb == 1, '', 's')}</span>", collapse = FALSE), collapse = ", "), qq(" within one week. Summary table was generated at @{format(Sys.time())}."))
	    txt = paste0(txt, " ", as.character(actionLink("job_stats", "See job stats")))
        HTML(txt)
    })
    
	observeEvent(input$select_job, {

        job_tb = job_summary_table()

	    job_id = gsub(" ", "", input$select_job)
        job_name = job_tb$JOB_NAME[job_tb$JOBID == job_id]
        job_status = job_tb$STAT[job_tb$JOBID == job_id]
	    message(qq("[@{format(Sys.time())}] Click 'job @{job_id}'"))

	    output$job_log = renderUI({
            if(job_status %in% c("RUN", "PEND") || is.null(ENV$job_log[[job_id]])) {
    	        showNotification("Fetching job log...", duration = 5)
    	        message(qq("[@{format(Sys.time())}] Fetching job log for '@{job_id} <@{job_name}> @{job_status}'"))
    	        log = job_log(job_id, print = FALSE)
                log = gsub("<", "&lt;", log)
                log = gsub(">", "&gt;", log)

                log = reformat_log(log, job_status)
                if(!job_status %in% c("RUN", "PEND")) {
                    ENV$job_log[[job_id]] = log
                }
            } else {
                log = ENV$job_log[[job_id]]
                message(qq("[@{format(Sys.time())}] Fetching job log for '@{job_id} <@{job_name}> @{job_status}' from cache"))
                
            }

	        html = paste(log, collapse = "\n")
            html = paste0("<pre>", html, "</pre>")
            HTML(html)
	    })

        output$job_env = renderUI({
            log = run_cmd(qq("bjobs -env @{job_id}"))
            names(log) = gsub("^(\\S+)=.*$", "\\1", log)

            log = gsub("^(\\w+)", "<b>\\1</b>", log)
            log = paste("<code>", log, "</code>")
            
            html = paste(log, collapse = "\n")
            html = paste0("<pre>", html, "</pre>")
             HTML(html)
            
        })

        output$job_script = renderUI({
            log = run_cmd(qq("bjobs -script @{job_id}"))
            i1 = grep("^# LSBATCH: User input", log)
            source = FALSE
            if(log[i1+1] == "( cat <<_USER_\\SCRIPT_") {
                i1 = i1 + 1
                source = TRUE
            }
            if(source) {
                i2 = which(log == "_USER_SCRIPT_")
                log = log[seq(i1+1, i2-1)]
            } else {
                i2 = grep("^# LSBATCH: End user input", log)
                log = log[seq(i1+1, i2-1-2)]
            }
            
            log = paste("<code>", log, "</code>")
            html = paste(log, collapse = "\n")
            html = paste0("<pre>", html, "</pre>")
            HTML(html)
        })

	    showModal(modalDialog(
	        title = qq("Job log (@{job_id} <@{job_name}>, @{job_status})"),
            if(job_status %in% c("EXIT", "unknown")) {
                tags$p(HTML(qq("<a style='color:#337ab7;text-decoration:underline;cursor:pointer;' onclick='Shiny.onInputChange(\"select_dep\", 0);Shiny.onInputChange(\"select_dep\", @{job_id})'>Show job dependencies</a>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a style='color:#337ab7;text-decoration:underline;cursor:pointer;' onclick='$(\"#rerun_pipeline_output\").show();rerun_pipeline(\"@{job_id}\")'>Click here to rerun the pipeline (DONE jobs will be skipped).</a>")))
            } else {
                tags$p(HTML(qq("<a style='color:#337ab7;text-decoration:underline;cursor:pointer;' onclick='Shiny.onInputChange(\"select_dep\", 0);Shiny.onInputChange(\"select_dep\", @{job_id})'>Show job dependencies</a>")))
            },
            if(job_status %in% c("EXIT", "unknown")) {
                HTML("<pre class='shiny-text-output noplaceholder' id='rerun_pipeline_output' style='display:hidden'></pre>")
            } else {
                NULL
            },
	        tabsetPanel(type= "tabs",
                tabPanel("Job log", uiOutput("job_log")),
                tabPanel("Script", uiOutput("job_script")),
                tabPanel("Env variables", uiOutput("job_env"))
            ),
	        footer = actionButton("close_job_log", "Close"),
	        easyClose = FALSE,
	        size = "l"
	    ))
	})

    observeEvent(input$select_dep, {

        job_tb = job_summary_table()

        job_id = gsub(" ", "", input$select_dep)
        job_name = job_tb$JOB_NAME[job_tb$JOBID == job_id]
        job_status = job_tb$STAT[job_tb$JOBID == job_id]
        message(qq("[@{format(Sys.time())}] Click dependency for 'job @{job_id}'"))

        # test whether the pipeline can be rerun
        g = job_dependency_igraph(job_id, job_tb)
        all_job_ids = V(g)$name

        job_tb2 = job_tb[job_tb$JOBID %in% all_job_ids, , drop = FALSE]
        rerunable = FALSE
        if(length(all_job_ids) == 1) {
            if(job_tb2$STAT[1] %in% c("DONE", "EXIT", "unknown")) {
                rerunable = TRUE
            }
        }
        if(all(job_tb2$STAT %in% c("DONE", "EXIT", "unknown"))) {
            rerunable = TRUE
        }
        if(all(job_tb2$STAT == "DONE")) {
            rerunable = FALSE
        }

        output$dependency_plot = renderGrViz({
            showNotification("Generating job dependency tree...", duration = 5)
            message(qq("[@{format(Sys.time())}] Generating job dependency tree '@{job_id} <@{job_name}> @{job_status}'"))
            dot = job_dependency_dot(job_id, job_tb, use_label = TRUE)
            grViz(dot, width = 868, height = 600)
        })
        
        showModal(modalDialog(
            title = qq("Job dependency (@{job_id} <@{job_name}>)"),
            if(rerunable) tags$p(HTML(qq("<a style='color:#337ab7;text-decoration:underline;cursor:pointer;' onclick='$(\"#rerun_pipeline_output\").show();rerun_pipeline(\"@{job_id}\")'>Some jobs are failed, click here to rerun the pipeline (DONE jobs will be skipped).</a>"))) else NULL,
            if(rerunable) HTML("<pre class='shiny-text-output noplaceholder' id='rerun_pipeline_output' style='display:hidden'></pre>") else NULL,
            HTML("<p id='dep_info'>If you cannot click on the nodes, move your mouse outside of the diagram and then move in.</p>"),
            grVizOutput("dependency_plot", width = "868px", height = "600px"),
            tags$script("
                $('#dependency_plot').on('mouseenter', function(e) {
                    gviz_add_events()
                })"
            ),
            HTML("<p><center><b>Legend:</b> <span class='RUN'>RUN</span> <span class='PEND'>PEND</span> <span class='DONE'>DONE</span> <span class='EXIT'>EXIT</span> <span class='unknown'>unknown</span></center></p>"),
            footer = actionButton("close_job_dep", "Close"),
            easyClose = FALSE,
            size = "l"
        ))
    })

    observeEvent(input$close_job_log, {
        output$job_log = renderUI({pre("loading...")})
        output$job_env = renderUI({pre("loading...")})
        output$job_script = renderUI({pre("loading...")})
        removeModal()
    })

    observeEvent(input$close_job_dep, {
        removeModal()
    })

    observeEvent(input$kill_job_id, {
        
        job_id = input$kill_job_id
        job_id = gsub("@.*$", "", job_id)
        job_id = strsplit(job_id, ";")[[1]]
        tb = job_summary_table()

        if(length(job_id) == 0) {
            output$kill_job_info = renderText({
                "No job was selected."
            })
            showModal(modal_kill_nothing)
        } else {
            tb = tb[tb$JOBID %in% job_id, , drop = FALSE]
            killing_job_id = tb$JOBID[tb$STAT %in% c("RUN", "PEND")]
            if(length(killing_job_id) == 0) {
                output$kill_job_info = renderText({
                    "No RUN/PEND job was selected."
                })
                showModal(modal_kill_nothing)
            } else {
                output$kill_job_info = renderText({
                    paste0("Kill following jobs? ", paste(killing_job_id, collapse = ","))
                })
                
                m = structure(names = paste0(tb$JOBID, " <", tb$JOB_NAME, ">"), tb$JOBID)
                
                showModal(modalDialog(
                    title = "Kill jobs",
                    checkboxGroupInput("killing_job_id", "Kill following jobs?", m[m %in% killing_job_id], selected = killing_job_id),
                    HTML("<pre class='shiny-text-output noplaceholder' id='kill_job_output' style='display:hidden'></pre>"),
                    footer = list(
                        actionButton("kill_job_cancel", "Cancel"),
                        HTML('<button id="kill_job_confirm" type="button" class="btn btn-default action-button">Kill</button>')
                    )
                ))
            }
        }
    })
    
    observeEvent(input$kill_no_job, {
        removeModal()
    })
    
    observeEvent(input$kill_job_confirm, {
        killing_job_id = input$killing_job_id
        showNotification(qq("Kill jobs"), duration = 5)
        withCallingHandlers({
            shinyjs::html("kill_job_output", "")
            for(id in killing_job_id) {
                message(qq("Kill job @{id}"))
                try(bkill(id), silent = TRUE)
            }
            for(sec in 10:1) {
                message(qq("Reload the app (sleep @{sec} seconds)"))
                Sys.sleep(1)
            }
        }, message = function(m) {
            shinyjs::html(id = "kill_job_output", html = m$message, add = TRUE)
        })
        session$reload()
    })
    
    observeEvent(input$kill_job_cancel, {
        removeModal()
    })

    observeEvent(input$job_stats, {

        job_tb = job_summary_table()

        message(qq("[@{format(Sys.time())}] query global job statistics"))

        output$job_barplot = renderPlot({
            showNotification("Generating job barplot...", duration = 2)
            message(qq("[@{format(Sys.time())}] generating job barplot"))
            bjobs_barplot(job_tb = job_tb)
        })

        output$job_timeline = renderPlot({
            showNotification("Generating job timeline...", duration = 2)
            message(qq("[@{format(Sys.time())}] generating job timeline"))
            bjobs_timeline(job_tb = job_tb)
        })
        
        showModal(modalDialog(
            title = qq("Job statistics"),
            p(h4("Number of jobs per day")),
            plotOutput("job_barplot", width = "760px", height = "600px"),
            p(h4("Duration of jobs")),
            plotOutput("job_timeline", width = "760px", height = "600px"),
            footer = actionButton("close_job_stats", "Close"),
            easyClose = FALSE,
            size = "l"
        ))
    })

    observeEvent(input$close_job_stats, {
        output$job_barplot = NULL
        output$job_timeline = NULL
        removeModal()
    })

    observeEvent(input$rerun_pipeline, {
        job_id = input$rerun_pipeline_job_id
        showNotification(qq("rerun pipeline by @{job_id}..."), duration = 5)

        withCallingHandlers({
            shinyjs::html("rerun_pipeline_output", "")
            pipeline_rerun(job_id)
            for(sec in 10:1) {
                message(qq("Reload the app (sleep @{sec} seconds)"))
                Sys.sleep(1)
            }
        },
        message = function(m) {
            shinyjs::html(id = "rerun_pipeline_output", html = m$message, add = TRUE)
        })

        session$reload()
    })
}

modal_kill_nothing = modalDialog(
    title = "Kill jobs",
    textOutput("kill_job_info"),
    footer =  actionButton("kill_no_job", "OK"),
    easyClose = TRUE
)

reformat_log = function(log, status) {
    if(status == "EXIT") {
        ind = which(grepl("\\berror\\b|\\bfail|Exited with exit code", log, ignore.case = TRUE) & !grepl("^#", log))
        if(length(ind)) {
            log[ind] = paste0("<span style='color:#e41a1c;font-weight:bold;text-decoration:underline;'>", log[ind], "</span>")
        }
    }
    if(status == "DONE") {
        ind = grep("Successfully completed", log, ignore.case = TRUE)
        if(length(ind)) {
            log[ind] = paste0("<span style='color:#377eb8;font-weight:bold;text-decoration:underline;'>", log[ind], "</span>")
        }
    }
    if(status == "RUN") {
        ind = grep("is still running", log)
        if(length(ind)) {
            log[ind] = paste0("<span style='color:darkgreen;font-weight:bold;text-decoration:underline;'>", log[ind], "</span>")
        }
    }
    if(status == "PEND") {
        ind = grep("pending", log)
        if(length(ind)) {
            log[ind] = paste0("<span style='color:#ff7f00;font-weight:bold;text-decoration:underline;'>", log[ind], "</span>")
        }
    }
    paste("<code>", log, "</code>")
}

shinyApp(ui, server)

#' Job status by job ID or name
#'
#' @param job_name A single job name.
#' @param job_id A single job ID.
#'
#' @returns A vector of job status, with job IDs are names.
#' @export
#' @rdname job_status
#' @examples
#' \dontrun{
#' job_status_by_name("example")
#' job_status_by_id(123456)
#' }
job_status_by_name = function(job_name) {

    ln = run_cmd(qq("bjobs -a -o 'jobid stat delimiter=\",\"' -J @{job_name} 2>&1"), print = FALSE)

    if(length(ln) == 1) {
        return("MISSING")
    }

    ln = ln[-1]
    lt = strsplit(ln, ",")

    structure(sapply(lt, function(x) x[2]), names = sapply(lt, function(x) x[1]))
}

#' @export
#' @rdname job_status
job_status_by_id = function(job_id) {

    if(missing(job_id)) {
        job_id = ""
    }

    ln = run_cmd(qq("bjobs -a -o 'jobid stat delimiter=\",\"' @{job_id} 2>&1"), print = FALSE)

    if(length(ln) == 1) {
        return("MISSING")
    }

    ln = ln[-1]
    lt = strsplit(ln, ",")

    structure(sapply(lt, function(x) x[2]), names = sapply(lt, function(x) x[1]))
}

job_output_file_by_id = function(job_id) {
    
    if(missing(job_id)) {
        job_id = ""
    }

    ln = run_cmd(qq("bjobs -a -o 'jobid output_file delimiter=\",\"' @{job_id} 2>&1"), print = FALSE)

    if(length(ln) == 1) {
        return(NA)
    }

    ln = ln[-1]
    lt = strsplit(ln, ",")

    structure(sapply(lt, function(x) x[2]), names = sapply(lt, function(x) x[1]))
}

job_output_file_by_name = function(job_name) {
    ln = run_cmd(qq("bjobs -a -o 'jobid output_file delimiter=\",\"' -J \"@{job_name}\" 2>&1"), print = FALSE)

    if(length(ln) == 1) {
        return(NA)
    }

    ln = ln[-1]
    lt = strsplit(ln, ",")

    structure(sapply(lt, function(x) x[2]), names = sapply(lt, function(x) x[1]))
}

job_attached_vars_by_id = function(job_id) {

    if(missing(job_id)) {
        job_id = ""
    }

    ln = run_cmd(qq("bjobs -a -o 'job_description' @{job_id} 2>&1"), print = FALSE)

    if(length(ln) == 1) {
        return(NA)
    }

    ln = ln[-1]
    lt = strsplit(ln, ",")

    jid = sapply(lt, function(x) x[1])
    jd = sapply(lt, function(x) x[2])
    jd[jd == "-"] = NA

    name_uid = gsub("^.*R_BSUB_NAME_UID=(.*),?.*$", "\\1", jid)
    temp_dir = gsub("^.*R_BSUB_TEMP_DIR=(.*)$", "\\1", jd)

    names(name_uid) = job_id
    names(temp_dir) = job_id

    list(name_uid = name_uid, temp_dir = temp_dir)
}

job_attached_vars_by_name = function(job_name) {

    ln = run_cmd(qq("bjobs -a -o 'job_description jobid delimiter=\",\"' -J @{job_name} 2>&1"), print = FALSE)

    if(length(ln) == 1) {
        return(NA)
    }

    ln = ln[-1]
    lt = strsplit(ln, ",")

    job_id = sapply(lt, function(x) x[3])
    jid = sapply(lt, function(x) x[1])
    jd = sapply(lt, function(x) x[2])
    jd[jd == "-"] = NA

    name_uid = gsub("^.*R_BSUB_NAME_UID=(.*),?.*$", "\\1", jid)
    temp_dir = gsub("^.*R_BSUB_TEMP_DIR=(.*)$", "\\1", jd)

    names(name_uid) = job_id
    names(temp_dir) = job_id

    list(name_uid = name_uid, temp_dir = temp_dir)
}

#' Clear temporary dir
#' 
#' @param ask Whether to promote.
#'
#' @details
#' The temporary files might be used by the running/pending jobs. Deleting them would affect some jobs.
#' You should better delete them after all jobs are done.
#'
#' @export
#' @examples
#' \dontrun{
#' list_temp_files()
#' }
#' 
list_temp_files = function() {

    if(!under_same_file_system()) {
        stop("Temporary files can only be detected on the same file system as submission nodes.")
    }

    attached_vars = job_attached_vars_by_id()
    temp_dir = attached_vars$temp_dir
    temp_dir = temp_dir[!is.na(temp_dir)]

    files = unlist(lapply(temp_dir, function(x) {
        list.files(x, full.names = TRUE)
    }))

    unname(files)
}

#' @rdname list_temp_files
#' @export
remove_temp_files = function(ask = TRUE) {

    if(!under_same_file_system()) {
        stop("Temporary files can only be detected on the same file system as submission nodes.")
    }

    files = list_temp_files()

    if(length(files) == 0) {
        if(ask) qqcat("All temporary directories are empty.\n")
        return(invisible(NULL))
    }
    if(!ask) {
        if(length(files)) {
            file.remove(files)
            return(invisible(NULL))
        }
    }

    file_types = gsub("^.*\\.([^.]+)$", "\\1", files)
    tb = table(file_types)
    job_tb = bjobs(status = "all", print = FALSE)
    if(any(job_tb$JOB_STAT %in% c("RUN", "PEND"))) {
        cat("There are still running/pending jobs. Deleting temporary files might affect some of the jobs.\n")
    }
    cat(qq("There @{ifelse(length(files) > 1, 'are', 'is')} "), qq("@{tb} .@{names(tb)} file@{ifelse(tb > 1, 's', '')}, "), "delete all? [y|n|s] ", sep = "")
    while(1) {
        answer = readline()
        if(answer == "y" || answer == "Y") {
            file.remove(files)
            break
        } else if(answer == "n" || answer == "N") {
            break
        } else if(answer == "s" || answer == "S") {
            for(nm in names(tb)) {
                qqcat("Remove all @{tb[nm]} .@{nm} file@{ifelse(tb[nm] > 1, 's', '')}? [y|n] ")
                while(1) {
                    answer2 = readline()
                    if(answer2 == "y" || answer2 == "Y") {
                        file.remove(files[file_types == nm])
                        break
                    } else if(answer2 == "n" || answer2 == "N") {
                        break
                    } else {
                        qqcat("Remove all @{tb[nm]} .@{nm} file@{ifelse(tb[nm] > 1, 's', '')}? [y|n] ")
                    }
                }
                
            }
            break
        } else {
            cat(qq("There @{ifelse(length(files) > 1, 'are', 'is')} "), qq("@{tb} .@{names(tb)} file@{ifelse(tb > 1, 's', '')}, "), "delete all? [y|n|s] ", sep = "")
        }
    }
    return(invisible(NULL))
}

#' Check whether there are dump files
#' 
#' @param print Whether to print messages.
#'
#' @details
#' For the failed jobs, LSF cluster might generate a core dump file and R might generate a `.RDataTmp` file.
#'
#' Note if you manually set working directory in your R code/script, the R dump file can be not caught.
#'
#' @returns A vector of file names.
#' @export
#' @examples
#' \dontrun{
#' list_dump_files()
#' }
list_dump_files = function(print = TRUE) {

    if(!under_same_file_system()) {
        stop("Dump files can only be detected on the same file system as submission nodes.")
    }

    job_tb = bjobs(status = "all", print = FALSE)
    wd = job_tb$EXEC_CWD
    wd = wd[wd != "-"]
    wd = unique(wd)

    dump_files = NULL
    for(w in wd) {
        if(print) {
            qqcat("checking @{w}\n")
        }
        dump_files = c(dump_files, list.files(path = w, pattern = "^core.\\d$", all.files = TRUE, full.names = TRUE))
        dump_files = c(dump_files, list.files(path = w, pattern = "^\\.RDataTmp", all.files = TRUE, full.names = TRUE))
    }

    return(dump_files)
}

#' Connect to submisstion via ssh
#'
#' @param ... Pass to [`ssh::ssh_connect()`].
#' 
#' @importFrom ssh ssh_session_info
#' @rdname ssh
#' @export
#' @examples
#' # ssh is automatically connected. To manually connect ssh, run:
#' \dontrun{
#' ssh_connect()
#' ssh_disconnect()
#' }
#' # where the user name is the one you set in `bsub_opt$user` and
#' # the node is the one you set in `bsub_opt$login_node`.
ssh_connect = function(...) {

	if(!is.null(bsub_opt$ssh_session)) {
		if(ssh::ssh_session_info(bsub_opt$ssh_session)$connected) {
			message("ssh is already connected.")
			return(invisible(NULL))
		}
	}

	submission_node = bsub_opt$submission_node
	login_node = bsub_opt$login_node
	if(length(login_node) == 0) login_node = submission_node

	if(length(login_node) == 0) {
		stop("bsub_opt$submission_node is not defined. No node to connect.")
	}

	if(Sys.info()["nodename"] %in% login_node) {
		message("already on submission/login node. No need to reconnect.")
	}

	user = bsub_opt$user

	for(i in seq_along(login_node)) {
		message(qq("establish ssh connection to @{user}@@{login_node[i]}"))
		oe = try(session <- ssh::ssh_connect(paste0(user, "@", login_node[i]), ...))

		if(!inherits(oe, "try-error")) {
			bsub_opt$ssh_session = session
			break
		}
	}

	if(is.null(bsub_opt$ssh_session)) {
		stop("cannot connect to submission nodes.")
	}
}



#' @rdname ssh
#' @export
ssh_disconnect = function() {
	if(!is.null(bsub_opt$ssh_session)) {
		ssh::ssh_disconnect(bsub_opt$ssh_session)
	}
	bsub_opt$ssh_session = NULL
}

ssh_exec = function(cmd) {

	ssh_validate()

	if(bsub_opt$verbose) cat("========== ssh commands =============", bsub_opt$ssh_envir, cmd, sep = "\n")

	ln = ""
	con = textConnection("ln", "w", local = TRUE)
    status = ssh::ssh_exec_wait(bsub_opt$ssh_session, command = c(
        bsub_opt$ssh_envir,
        cmd
    ), std_out = con, std_err = con)
    close(con)
    if(status) {
    	stop(paste(c("You have an error when executing remote commands", ln), collapse = "\n"))
    }
    return(ln)
}

ssh_validate = function() {
	if(is.null(bsub_opt$ssh_session)) {
		ssh_connect()
	} else {
		session = bsub_opt$ssh_session
		if(ssh::ssh_session_info(session)$connected) {
			return(NULL)
		} else {
			bsub_opt$ssh_session = NULL
			ssh_connect()
		}
	}
}

 



.onAttach = function(libname, pkgname) {

	version = utils::packageDescription(pkgname, fields = "Version")

	msg = NULL
	msg = c(msg, "==================================================================")
	msg = c(msg, paste0("bsub version ", version))
	msg = c(msg, "Github page: https://github.com/jokergoo/bsub")
	msg = c(msg, "")

	msg = c(msg, "- submit R code: `bsub_chunk()`")
	msg = c(msg, "- submit R script: `bsub_script()`")
	msg = c(msg, "- submit shell commands: `bsub_cmd()`")
	msg = c(msg, "- kill jobs: `bkill()`")
	msg = c(msg, "- view job summary: `bjobs`/`brecent`/`bjobs_running`/")
	msg = c(msg, "                    `bjobs_pending`/`bjobs_done`/`bjobs_exit`")
	msg = c(msg, "- view job log: `job_log()`")
	msg = c(msg, "- interactive job monitor: `monitor()`")

	msg = c(msg, "")
	msg = c(msg, "`bsub_chunk()`/`bsub_script()`/`bsub_cmd()` should only be")
	msg = c(msg, "applied on the node that has the same file system as the computing")
	msg = c(msg, "nodes. Other functions for monitoring and cleaning jobs can be")
	msg = c(msg, "applied on any computer.")

	msg = c(msg, "==================================================================")

	packageStartupMessage(paste(msg, collapse = "\n"))
}


.onDetach = function(libpath) {
	if(!is.null(bsub_opt$ssh_session)) ssh_disconnect()
}

.onUnload = function(libpath) {
	if(!is.null(bsub_opt$ssh_session)) ssh_disconnect()
}

if(grepl("odcf|w610", Sys.info()["nodename"])) {
	config_odcf()
}

#' Job dependencies
#'
#' @param job_id A single job ID.
#' @param job_tb A data frame from [`bjobs()`]. Internally used.
#' @param ... Pass to [`DiagrammeR::grViz()`], such as the size of the html widget.
#'
#' @returns
#' `job_dependency_all()` returns a list that contains three elements:
#'
#' - `dep_mat`: a two column matrix containing dependencies from parents to children.
#' - `id2name`: a named vector containing mapping from job IDs to job names.
#' - `id2stat`: a named vector containing mapping from job IDs to job status.
#'
#' `job_dependency_igraph()` returns a [`igraph::igraph`] object which contains a dependency
#' graph induced by the input job ID.
#' 
#' `job_dependency_dot()` returns a DOT code for GraphViz visualization.
#' 
#' `job_dependency_diagram()` makes a HTML-based dependency diagram.
#' 
#' @rdname dependency
#' @export
#' @examples
#' \dontrun{
#' job1 = random_job()
#' job2 = random_job()
#' job3 = random_job(dependency = c(job1, job2))
#' 
#' job_dependency_all()
#' job_dependency_igraph(job3)
#' cat(job_dependency_dot(job3))
#' job_dependency_diagram(job3)
#' }
job_dependency_all = function(job_tb = NULL) {

    if(is.null(job_tb)) job_tb = bjobs(status = "all", print = FALSE)
    
    job_tb = job_tb[, c("JOBID", "STAT", "JOB_NAME", "DEPENDENCY"), drop = FALSE]

    id2name = structure(job_tb$JOB_NAME, names = job_tb$JOBID)
    id2stat = structure(as.character(job_tb$STAT), names = job_tb$JOBID)

    job_tb$DEPENDENCY = gsub("\\w+\\((.*?)\\)", "\\1", job_tb$DEPENDENCY)
    job_tb2 = job_tb[job_tb$DEPENDENCY != "-", , drop = FALSE]
    
    if(nrow(job_tb2) == 0) {
        dep_mat = cbind(parent = character(0),
                     child = character(0))

        return(list(dep_mat = dep_mat, id2name = id2name, id2stat = id2stat))
    }
    
    dep = lapply(strsplit(job_tb2$DEPENDENCY, " && "), function(x) {
        gsub(" ", "", gsub("^done\\( (\\d+) \\)$", "\\1", x))
    })

    n = sapply(dep, length)
    dep_mat = cbind(parent = as.character(unlist(dep)),
                    child = as.character(rep(job_tb2$JOBID, times = n)))

    return(list(dep_mat = dep_mat, id2name = id2name, id2stat = id2stat))
}

#' @rdname dependency
#' @export
#' @import igraph
job_dependency_igraph = function(job_id, job_tb = NULL) {
   
    job_id = as.character(job_id)
    job_dep = job_dependency_all(job_tb = job_tb)

    if(!job_id %in% job_dep$dep_mat) {
        g = induced_subgraph( make_graph(c(job_id, "foo")), vids = 1)
    } else {

        g = graph.edgelist(job_dep$dep_mat)
        g2 = graph.edgelist(job_dep$dep_mat, directed = FALSE)

        dist = distances(g2, v = job_id)

        # node in the connected sub-graph
        nodes = names(which(apply(dist, 2, function(x) any(is.finite(x)))))

        g = induced_subgraph(g, nodes)

    }
    V(g)$label = job_dep$id2name[V(g)$name]
    V(g)$status = job_dep$id2stat[V(g)$name]
    g
}

#' @param use_label Whether to use job names on the diagram?
#' @param label_width Max number of characters for wrapping the label into lines.
#' @rdname dependency
#' @importFrom grDevices col2rgb rgb
#' @export
job_dependency_dot = function(job_id, job_tb = NULL, use_label = FALSE, label_width = 15) {
        
    job_id = as.character(job_id)
    g = job_dependency_igraph(job_id, job_tb)

    ## the dot code
    color = STATUS_COL[V(g)$status]
    color = rgb(t(col2rgb(color)/255))
    fontsize = 10
    fontcolor = "black"
    penwidth = rep(1, length(color))
    penwidth[V(g)$name == job_id] = 3

    if(use_label) {
        label = V(g)$label
        label = gsub("_", " ", label)
        label = sapply(label, function(x) {
            paste(strwrap(x, width = label_width), collapse = "\\n")
        })
        label = paste0(label, "\\n", "(", V(g)$name, ")")
        nodes = paste0(
            qq("  node [fontname=Helvetical]\n"),
            qq("  \"@{V(g)$name}\" [shape=box, color=\"@{color}\", fillcolor=\"@{color}40\", style=filled, penwidth=@{penwidth}, fontsize=@{fontsize}, fontcolor=\"@{fontcolor}\", label=\"@{label}\", fontname=\"Arial\", tooltip=\"@{V(g)$name}\"];\n", collapse = TRUE)
        )
    } else {
        nodes = paste0(
            qq("  node [fontname=Helvetical]\n"),
            qq("  \"@{V(g)$name}\" [color=\"@{color}\", fillcolor=\"@{color}40\", style=filled, penwidth=@{penwidth}, fontsize=@{fontsize}, fontcolor=\"@{fontcolor}\", fontname=\"Arial\", tooltip=\"@{V(g)$label}\"];\n", collapse = TRUE)
        )
    }

    edgelist = get.edgelist(g)
    if(nrow(edgelist)) {
        edges = paste0(
            "  edge []\n",
            qq("  \"@{edgelist[, 1]}\" -> \"@{edgelist[, 2]}\";\n", collapse = TRUE)
        )
    } else {
        edges = ""
    }

    DOT = paste0(
        "digraph {\n",
        "  graph [overlap = true]\n",
        "\n",
        nodes,
        "\n",
        edges,
        "}\n"
    )

    DOT
}

#' @rdname dependency
#' @export
job_dependency_diagram = function(job_id, job_tb = NULL, use_label = FALSE, label_width = 15, ...) {

    job_id = as.character(job_id)
    dot = job_dependency_dot(job_id, job_tb = job_tb, use_label = use_label, label_width = label_width)
    DiagrammeR::grViz(dot, ...)
}

#' Obtain Job log
#'
#' @param job_id The job id. It can be a single job or a vector of job ids.
#' @param print Whether to print the log message to the terminal.
#' @param n_line Number of last lines for each job to show when multiple jobs are queried.
#' 
#' @returns The log messages as a vector.
#' @import clisymbols
#' @export
#' @examples
#' \dontrun{
#' # a single job
#' job_id = 1234567  # job ids can be get from `bjobs`
#' job_log(job_id)
#' # multiple jobs
#' job_id = c(10000000, 10000001, 10000002) 
#' job_log(job_id)  # by  default last 10 lines for each job are printed
#' job_log(job_id, n_line = 20) # print last 20 lines for each job
#' # logs for all running jobs
#' job_log()
#' }
job_log = function(job_id, print = TRUE, n_line = 10) {
    
    tb = bjobs(print = FALSE, status = "all")
    
    if(missing(job_id)) {
        tb = bjobs(print = FALSE)
    
        if(is.null(tb)) {
            txt = ("No running job")
            if(print) cat(txt, sep = "\n")
            return(invisible(txt))
        }
        return(job_log(tb[, 1], print = print, n_line = n_line))
    }

    job_id = as.character(job_id)

    if(!all(grepl("^\\d+$", job_id))) {
        txt = qq("Job ID should be all digits.")
        if(print) cat(txt, sep = "\n")
        return(invisible(txt))
    }

    if(length(job_id) > 1) {
        txt2 = NULL
        for(id in job_id) {
            if(print) qqcat("retrieve log for job @{id}\n")
            txt = job_log(id, print = FALSE)
            if(length(txt) > n_line) {
                txt2 = c(txt2, "\n", paste0(strrep(symbol$double_line, 10), qq(" log for job @{id}, last @{n_line} lines "), strrep(symbol$double_line, 10)))
                txt2 = c(txt2, txt[seq(length(txt) - n_line + 1, length(txt))])
            } else {
                txt2 = c(txt2, "\n", paste0(strrep(symbol$double_line, 10), qq(" log for job @{id} "), strrep(symbol$double_line, 10)))
                txt2 = c(txt2, txt)
            }
        }
        if(print) {
            cat(txt2, sep = "\n")
        }
        return(invisible(txt2))
    }

    # if the job with job_id is not the newest one with the same job name
    job_name = tb$JOB_NAME[tb$JOBID == job_id]
    job_wd = tb$EXEC_CWD[tb$JOBID == job_id]  
    job_queue = tb$QUEUE[tb$JOBID == job_id]  

    if(length(job_name) == 0) {
        txt = qq("Job @{job_id} does not exist.")
        if(print) cat(txt, sep = "\n")
        return(invisible(txt))
    }

    if(job_queue == "interactive") {
        txt = qq("Job @{job_id} is an interactive job and has no log.")
        if(print) cat(txt, sep = "\n")
        return(invisible(txt))
    }

    # tb_subset = tb[tb$JOB_NAME == job_name & tb$EXEC_CWD == job_wd, , drop = FALSE]
    # nrr = nrow(tb_subset)
    # if(nrr > 1) {
    #     tb_subset = tb_subset[order(tb_subset$JOBID), , drop = FALSE]
    #     if(tb_subset$JOBID[nrr] != job_id) {
    #         txt = qq("Log file for job @{job_id} has been overwriten by job @{tb_subset$JOBID[nrr]} which is the newest job with the same job name '@{tb_subset$JOB_NAME[nrr]}'.")
    #         txt = strwrap(txt)
    #         if(print) cat(txt, sep = "\n")
    #         return(invisible(txt))
    #     }
    # }

    ln = run_cmd(qq("bjobs -o \"stat output_file\" @{job_id} 2>&1"), print = FALSE)

    if(length(ln) == 1) {
        status = "MISSING"
    }

    ln = ln[-1]  # remove header
    status = gsub("\\s.*$", "", ln)
    output_file = gsub("^\\S+\\s+", "", ln)

    if(status == "MISSING") {
        txt = qq("Cannot find output file for job @{job_id}.")
        if(print) cat(txt, sep = "\n")
        return(invisible(txt))
    } else if(status == "PEND") {
        txt = qq("Job @{job_id} is still pending.")
        if(print) cat(txt, sep = "\n")
        return(invisible(txt))
    } else if(status == "RUN") {
        # bpeek is slow, we first directly check the temporary output file
        user = bsub_opt$user

        # the temporary job dir
        ln = run_cmd("bparams -a", print = FALSE)
        ind = grep("JOB_SPOOL_DIR", ln)
        if(length(ind)) {
            job_temp_dir = gsub("^\\s*JOB_SPOOL_DIR = ", "", ln[ind])
            job_temp_dir = gsub("/%U$", "", job_temp_dir)
            job_temp_dir = qq("@{job_temp_dir}/@{user}")
        } else {
            remote_home = run_cmd("echo $HOME", print = FALSE)
            job_temp_dir = qq("@{remote_home}/.lsbatch")
        }

        no_file_flag = FALSE
        # check running jobs
        if(on_submission_node()) {
            file = list.files(path = job_temp_dir, pattern = paste0("\\.", job_id,, "\\.out"), full.names = TRUE)

            if(length(file) == 0) {
                no_file_flag = TRUE
            } else {
                txt = readLines(file, warn = FALSE)
                txt = c(txt, paste0("\n", symbol$warning, qq(" job @{job_id} is still running.")))
                if(print) cat(txt, sep = "\n")
                return(invisible(txt))
            }
        } else {
            # if no such file, ssh_exec gives an error
            oe = try(ln <- ssh_exec(qq("ls @{job_temp_dir}/*.@{job_id}.out")), silent = TRUE)

            if(inherits(oe, "try-error")) {
                no_file_flag = TRUE
            } else if(length(ln) == 0) {
                no_file_flag = TRUE
            } else {
                txt = ssh_exec(qq("cat @{ln}"))
                txt = c(txt, paste0(symbol$warning, qq(" job @{job_id} is still running.")))
                if(print) cat(txt, sep = "\n")
                return(invisible(txt))
            }
        }

        if(no_file_flag) {
            ln = run_cmd(qq("bpeek @{job_id}"), print = FALSE)[-1]
            if(length(ln) == 0) {
                txt = qq("Cannot find output file for job @{job_id}.")
                if(print) cat(txt, sep = "\n")
                return(invisible(txt))
            } else {
                txt = ln
                txt = c(txt, paste0(symbol$warning, qq("\n job @{job_id} is still running. "), symbol$warning))
                if(print) cat(txt, sep = "\n")
                return(invisible(txt))
            }
        }
    } else {
        no_file_flag = FALSE
        if(on_submission_node()) {
            
            if(!file.exists(output_file)) {
                no_file_flag = TRUE
            } else {
                txt = readLines(output_file, warn = FALSE)
                txt = extract_job_log(txt, job_id)
                if(print) cat(txt, sep = "\n")
                return(invisible(txt))
            }
        } else {
            # if no such file, ssh_exec gives an error
            oe = try(ln <- ssh_exec(qq("ls @{output_file}")), silent = TRUE)

            if(inherits(oe, "try-error")) {
                no_file_flag = TRUE
            } else if(length(ln) == 0) {
                no_file_flag = TRUE
            } else {
                txt = ssh_exec(qq("cat @{ln}"))
                txt = extract_job_log(txt, job_id)
                if(print) cat(txt, sep = "\n")
                return(invisible(txt))
            }
        }

        if(no_file_flag) {
            txt = qq("Cannot find output file for job @{job_id}.")
            if(print) cat(txt, sep = "\n")
            return(invisible(txt))
        }
    }
}


# it is possible multiple job logs are in one out file
extract_job_log = function(ln, job_id) {
    n = length(ln)
    ind = grep("^Sender: LSF System", ln)
    all_ids = gsub("^Subject: Job (\\d+):.*$", "\\1", ln[ind+1])
    k = which(all_ids == job_id)
    nk = length(k)

    if(k == nk) {
        ln[seq(ind[k], n)]
    } else {
        ln[seq(ind[k], ind[k+1]-1)]
    }
}

#' Retrieve saved variable
#'
#' @param job_id A single job ID.
#' @param job_name A single job name. Since jobs may have the same names, the most recent job is selected.
#' @param wait Seconds to wait until the job is finished.
#'
#' @details
#' It retrieves the saved variable in [`bsub_chunk()`] when `save_rds = TRUE` is set.
#'
#' @returns The retrieved object.
#' @export
#' @examples
#' \dontrun{
#' job_id = bsub_chunk(name = "example", save_var = TRUE,
#' {
#'     Sys.sleep(10)
#'     1+1
#' })
#' retrieve_var(job_id)
#' }
retrieve_var = function(job_id, job_name = NULL, wait = 30) {

    if(!under_same_file_system()) {
        stop("Variable can only be retrieved on the same file system as submission node.")
    }

    if(!missing(job_id)) {
        output_dir = dirname(job_output_file_by_id(job_id))
        attached_vars = job_attached_vars_by_id(job_id)
        name_uid = attached_vars$name_uid
        temp_dir = attached_vars$temp_dir
    } else {
        output_file = job_output_file_by_name(job_name)
        if(length(output_file) > 1) {
            i = which.max(as.numeric(names(output_file)))
            message_wrap(qq("Found @{length(output_file)} jobs with name '@{job_name}'. Use the variable from the most recent job <@{names(output_file)[i]}>. If this is not what you want, please specify argument `job_id = ...`."))
        } else {
            i = 1
        }
        attached_vars = job_attached_vars_by_name(job_name)
        name_uid = attached_vars$name_uid
        temp_dir = attached_vars$temp_dir

        job_id = names(temp_dir)[i]
        output_dir = dirname(output_file[i])
        name_uid = name_uid[i]
        temp_dir = temp_dir[i]
    }

    job_tb = bjobs(status = "all", print = FALSE, job_id = job_id)
    name = job_tb$JOB_NAME

    rds_file = qq("@{output_dir}/@{name_uid}_returned_var.rds")
    out_file = qq("@{output_dir}/@{name_uid}.out")
    flag_file = qq("@{output_dir}/@{name}.done")

    if(file.exists(flag_file)) {
        if(file.exists(out_file)) {
            ln = readLines(out_file)
            if(any(grepl("^Successfully completed", ln))) {
                if(file.exists(rds_file)) {
                    return(readRDS(rds_file))
                } else {
                    stop("Maybe you forget to set `save_var = TRUE` in `bsub_chunk()`?")
                }
            } else {
                stop("job failed.")
            }
        } else {
            stop("job failed.")
        }
    }

    status = job_status_by_id(job_id)
    if(status %in% c("RUN", "PEND")) {
        message(qq("job is running or pending, retry in @{wait} seconds."))
        Sys.sleep(wait)
        retrieve_var(job_id, wait = wait)
    } else if(status == "DONE") {
        retrieve_var(job_id)
    } else if(status == "MISSING") {
        stop(qq("cannot find the job '@{job_name}'."))
    } else {
        stop("job failed.")
    }
}

### format in JS ### 

#' Format file size
#' 
#' @param table Internally used.
#' @param columns Internally used.
#' @importFrom utils getFromNamespace
#' @export
formatFileSize = function(table, columns) {
	getFromNamespace("formatColumns", ns = "DT")(table, columns, tplFileSize)
}

formatTimeDiff = function(table, columns) {
	getFromNamespace("formatColumns", ns = "DT")(table, columns, tplTimeDiff)
}

formatNumber = function(table, columns) {
	getFromNamespace("formatColumns", ns = "DT")(table, columns, tplNumber)
}


# file size should be in bytes
tplFileSize = function(...) {
  "DTWidget.formatFileSize(data);"
}


# time diff should be in secs
tplTimeDiff = function(...) {
  "DTWidget.formatTimeDiff(data);"
}

tplNumber = function(...) {
  "DTWidget.formatNumber(data);"
}

### format in terminal table ###

convert_to_byte = function(x) {
    suppressWarnings(num <- as.numeric(gsub("^.*(\\d+)(\\.\\d+).*$", "\\1\\2", x)))
    v = ifelse(grepl("K", x), num*1024, ifelse(grepl("M", x), num*1024^2, ifelse(grepl("G", x), num*1024^3, x)))
    suppressWarnings(as.numeric(v))
}

format_mem = function(x) {
    gsub(" (.)bytes", "\\1b", x)
}

format_difftime = function(x, add_unit = FALSE) {
    units(x) = "hours"
    t = as.numeric(x)

    hour = floor(t)
    min = floor((t - hour)*60)
    
    l = is.na(x)
    if(add_unit) {
        txt = paste0(hour, "h", ifelse(min < 10, paste0("0", min), min), "m")
    } else {
        txt = paste0(hour, ":", ifelse(min < 10, paste0("0", min), min))
    }
    txt[l] = "-"
    txt[t == 0] = "-"
    txt
}


convert_to_POSIXlt = function(x) {

    if(is.null(bsub_opt$parse_time)) {

        if(any(grepl("^\\w+\\s+\\d+\\s+\\d+:\\d+$", x))) { # Dec 1 18:00
            t = as.POSIXlt(x, format = "%b %d %H:%M")
        } else if(any(grepl("^\\w+\\s+\\d+\\s+\\d+:\\d+:\\d+$", x))) { # Dec 1 18:00:00
            t = as.POSIXlt(x, format = "%b %d %H:%M:%S")
        } else {                                        # Dec 1 18:00:00 2019
            t = as.POSIXlt(x, format = "%b %d %H:%M:%S %Y")
        }
    } else {
        t = bsub_opt$parse_time(x)
    }

    if(any(is.na(t) & x != "-")) {
        stop_wrap(qq("Cannot convert time string (e.g. '@{x[which(is.na(t))[1]]}'') to a `POSIXlt` object. Please set a proper parsing function for `bsub_opt$parse_time`. See ?bsub_opt for more details."))
    }

    if(inherits(t, "POSIXct")) t = as.POSIXlt(t)

    current_t = as.POSIXlt(Sys.time())
    l = t$year > current_t$year
    l[is.na(l)] = FALSE
    if(any(l)) {
        t[l]$year = t[l]$year - 1
    }

    return(t)
}

#' Summary of jobs
#'
#' @param status Status of the jobs. Use "all" for all status
#' @param max Maximal number of recent jobs.
#' @param filter Regular expression on job names.
#' @param print Whether to print the table.
#' @param job_id A single job ID, internally used.
#'
#' @details
#' There is an additional column "RECENT" which is the order
#' for the job with the same name. 1 means the most recent job.
#'
#' You can directly type ``bjobs`` without parentheses which runs `bjobs` with defaults.
#'
#' - `brecent` shows the most recent.
#' - `bjobs_done` shows the "DONE" jobs.
#' - `bjobs_exit` shows the "EXIT" jobs.
#' - `bjobs_pending` shows the "PEND" jobs.
#' - `bjobs_running` shows the "RUN" jobs.
#'
#' @returns A data frame with selected job summaries.
#' @rdname bjobs
#' @importFrom utils read.csv
#' @export
#' @examples
#' \dontrun{
#' bjobs # this is the same as bjobs()
#' bjobs() # all running and pending jobs
#' bjobs(status = "all") # all jobs
#' bjobs(status = "RUN") # all running jobs, you can also use `bjobs_running`
#' bjobs(status = "PEND") # all pending jobs, you can also use `bjobs_pending`
#' bjobs(status = "DONE") # all done jobs, you can also use `bjobs_done`
#' bjobs(status = "EXIT") # all exit jobs, you can also use `bjobs_exit`
#' bjobs(status = "all", max = 20) # last 20 jobs
#' bjobs(status = "DONE", filter = "example") # done jobs with name '.*example.*'
#' }
bjobs = function(status = c("RUN", "PEND"), max = Inf, filter = NULL, print = TRUE, job_id = NULL) {

    if(is.null(job_id)) {
        cmd = "bjobs -a -o 'jobid stat job_name queue submit_time start_time finish_time runtimelimit nreq_slot mem max_mem dependency exec_cwd combined_resreq delimiter=\",\"' 2>&1"
    } else {
        cmd = qq("bjobs -a -o 'jobid stat job_name queue submit_time start_time finish_time runtimelimit nreq_slot mem max_mem dependency exec_cwd combined_resreq delimiter=\",\"' @{job_id} 2>&1")
    }
    ln = run_cmd(cmd, print = FALSE)
    
    if(length(ln) == 1) {
        cat(ln, "\n")
        return(invisible(NULL))
    }

    i = grep("^JOBID,STAT,JOB_NAME", ln)
    if(length(i) == 0) {
        cat("no output from bjobs\n")
        return(invisible(NULL))
    }

    ln = ln[seq(i, length(ln))]

    df = read.csv(textConnection(paste(ln, collapse = "\n")), stringsAsFactors = FALSE)
    df$STAT = factor(df$STAT)

    df$REQ_MEM = as.numeric(gsub("^.*mem=(.*?)\\].*$", "\\1", df$COMBINED_RESREQ))
    df$REQ_MEM = round(df$REQ_MEM/1024, 1)

    df$SUBMIT_TIME = convert_to_POSIXlt(df$SUBMIT_TIME)
    df$START_TIME = convert_to_POSIXlt(df$START_TIME)
    df$FINISH_TIME = convert_to_POSIXlt(df$FINISH_TIME)

    if(!is.null(bsub_opt$history_timestamp)) {
        l = df$FINISH_TIME > bsub_opt$history_timestamp
        df = df[l, , drop = FALSE]
    }

    # running/pending jobs
    df$TIME_PASSED = difftime(Sys.time(), df$START_TIME, units = "hours")
    l = !(df$STAT %in% c("RUN", "PEND"))
    # l[is.na(l)] = FALSE  # finish time is unavailable
    df$TIME_PASSED[l] = difftime(df$FINISH_TIME[l], df$START_TIME[l], units = "hours")
    df$TIME_LEFT = difftime(df$FINISH_TIME, Sys.time(), units = "hours")
    l = df$FINISH_TIME < Sys.time()
    l[is.na(l)] = TRUE
    df$TIME_LEFT[l] = NA

    df$JOBID = as.numeric(df$JOBID)
    df = df[order(df$JOBID), , drop = FALSE]
    tb = table(df$STAT)
    tb_today = table(df$STAT[as.numeric(difftime(Sys.time(), df$SUBMIT_TIME, units = "secs")) < 3600*24])

    recent = unlist(unname(tapply(df$JOBID, df$JOB_NAME, function(x) {
        structure(order(-x), names = x)
    }, simplify = FALSE)))
    df$RECENT = recent[as.character(df$JOBID)]

    if(! "all" %in% status) {
        df = df[df$STAT %in% toupper(status), , drop = FALSE]
    }
    if(!is.null(filter)) {
        df = df[grepl(filter, df$JOB_NAME), , drop = FALSE]
    }
    if(nrow(df)) {
        ind = sort((nrow(df):1)[1:min(c(nrow(df), max))])

        if(!print) {
            return(df[ind, , drop = FALSE])
        }

        df2 = format_summary_table(df[ind, , drop = FALSE])

        max_width = pmax(apply(df2, 2, function(x) max(nchar(x)+1)),
                         nchar(colnames(df2)) + 1)
        ow = getOption("width")
        options(width = sum(max_width) + 10)
        cat(strrep("=", sum(max_width)), "\n")
        print(df2, row.names = FALSE, right = FALSE, max = 99999)
        if(nrow(df2) > 20) {
            for(i in seq_len(ncol(df2))) {
                nm = colnames(df2)[i]
                cat(" ", nm, sep = "")
                cat(strrep(" ", max_width[i] - nchar(nm) - 1), sep = "")
            }
            cat("\n")
        }
        cat(strrep("=", sum(max_width)), "\n")
        cat(" ", paste(qq("@{tb} @{names(tb)} job@{ifelse(tb == 1, '', 's')}", collapse = FALSE), collapse = ", "), " within one week.\n", sep = "")
        cat(" ", paste(qq("@{tb_today} @{names(tb_today)} job@{ifelse(tb_today == 1, '', 's')}", collapse = FALSE), collapse = ", "), " in the last 24 hours.\n", sep = "")
        cat(" You can have more controls by `bjobs(status = ..., max = ..., filter = ...)`.\n")
        cat(" Use `brecent` to retrieve recent jobs from all status.\n")
        options(width = ow)

        return(invisible(df[ind, , drop = FALSE]))
    } else {
        if(!print) {
            return(NULL)
        }

        msg = paste0("status = '", paste(status, collapse = ", "), "'")
        if(!is.null(filter)) {
            msg = qq("@{msg}, filter = '@{filter}'")
        }
        qqcat("No job found (@{msg}).\n")
        cat("\n")
        cat(strrep("=", 78), "\n")
        cat(" ", paste(qq("@{tb} @{names(tb)} job@{ifelse(tb == 1, '', 's')}", collapse = FALSE), collapse = ", "), " within one week.\n", sep = "")
        cat(" ", paste(qq("@{tb_today} @{names(tb_today)} job@{ifelse(tb_today == 1, '', 's')}", collapse = FALSE), collapse = ", "), " in the last 24 hours.\n", sep = "")
        cat(" You can have more controls by `bjobs(status = ..., max = ..., filter = ...)`.\n")
        cat(" Use `brecent` to retrieve recent jobs from all status.\n")
        return(invisible(NULL))
    }
}
class(bjobs) = "bjobs"

#' @param fields Supported output fields, check \url{https://www.ibm.com/docs/en/spectrum-lsf/10.1.0?topic=information-customize-job-output}. 
#'        The value can be a vector of field names or a single string that already includes output fields separated by space.
#' @details
#' `bjobs_raw()` returns the table from the original `bsubs -a -o '...'` call.
#' @rdname bjobs
#' @export
bjobs_raw = function(fields = "jobid stat job_name queue") {

    if(length(fields) == 1) {
        fields = paste(fields, collapse = " ")
    }

    cmd = qq("bjobs -a -o '@{fields} delimiter=\",\"' 2>&1")
    ln = run_cmd(cmd, print = FALSE)
    
    # job done or exit
    if(length(ln) == 1) {
        cat(ln, "\n")
        return(invisible(NULL))
    }

    read.csv(textConnection(paste(ln, collapse = "\n")), stringsAsFactors = FALSE)
}

#' Clear job history
#'
#' @details
#' It sets a timestamp to only show jobs after it.
#' @export
bjobs_reset_timestamp = function() {
    bsub_opt$history_timestamp = Sys.time()
}


format_summary_table = function(df) {
    df2 = df[, c("JOBID", "STAT", "JOB_NAME", "RECENT","SUBMIT_TIME", "TIME_PASSED", "TIME_LEFT", "NREQ_SLOT", "MEM", "MAX_MEM")]

    df2$TIME_PASSED = format_difftime(df2$TIME_PASSED)
    df2$TIME_LEFT = format_difftime(df2$TIME_LEFT)
    df2$MAX_MEM = format_mem(df2$MAX_MEM)
    df2$MEM = format_mem(df2$MEM)

    if(all(df2$RECENT == 1)) {
        df2$RECENT = NULL
    }

    l = nchar(df2$JOB_NAME) > 50
    if(any(l)) {
        foo = substr(df2$JOB_NAME[l], 1, 48)
        foo = paste(foo, "..", sep = "")
        df2$JOB_NAME[l] = foo
    }

    return(df2)
}


#' Summary of jobs
#'
#' @param x A `bjobs` object.
#' @param ... other arguments.
#' 
#' @export
print.bjobs = function(x, ...) {
    x()
}


#' @rdname bjobs
#' @export
#' @examples
#' \dontrun{
#' brecent  # this is the same as `brecent()`
#' brecent() # last 20 jobs (from all status)
#' brecent(max = 50) # last 50 jobs
#' brecent(filter = "example") # last 20 jobs with name ".*example.*"
#' }
brecent = function(max = 20, filter = NULL) {
    bjobs(status = "all", max = max, filter = filter)
}
class(brecent) = "bjobs"

#' @rdname bjobs
#' @export
#' @examples
#' \dontrun{
#' bjobs_running  # this is the same as `bjobs_running()`
#' bjobs_running() # all running jobs
#' bjobs_running(max = 50) # last 50 running jobs
#' bjobs_running(filter = "example") # running jobs with name ".*example.*"
#' }
bjobs_running = function(max = Inf, filter = NULL) {
    bjobs(status = "RUN", max = max, filter = filter)
}
class(bjobs_running) = "bjobs"

#' @rdname bjobs
#' @export
#' @examples
#' \dontrun{
#' bjobs_pending  # this is the same as `bjobs_pending()`
#' bjobs_pending() # all pending jobs
#' bjobs_pending(max = 50) # last 50 pending jobs
#' bjobs_pending(filter = "example") # pending jobs with name ".*example.*"
#' }
bjobs_pending = function(max = Inf, filter = NULL) {
    bjobs(status = "PEND", max = max, filter = filter)
}
class(bjobs_pending) = "bjobs"

#' @rdname bjobs
#' @export
#' @examples
#' \dontrun{
#' bjobs_done  # this is the same as `bjobs_done()`
#' bjobs_done() # all done jobs
#' bjobs_done(max = 50) # last 50 done jobs
#' bjobs_done(filter = "example") # done jobs with name ".*example.*"
#' }
bjobs_done = function(max = Inf, filter = NULL) {
    bjobs(status = "DONE", max = max, filter = filter)
}
class(bjobs_done) = "bjobs"

#' @rdname bjobs
#' @export
#' @examples
#' \dontrun{
#' bjobs_exit  # this is the same as `bjobs_exit()`
#' bjobs_exit() # all exit jobs
#' bjobs_exit(max = 50) # last 50 exit jobs
#' bjobs_exit(filter = "example") # exit jobs with name ".*example.*"
#' }
bjobs_exit = function(max = Inf, filter = NULL) {
    bjobs(status = "EXIT", max = max, filter = filter)
}
class(bjobs_exit) = "bjobs"

#' Configure bsub global options
#' 
#' @param verbose Whether to print messages.
#' 
#' @details
#' It sets the submission nodes, user name and how to call `Rscript`.
#' @export
config_bsub = function(verbose = TRUE) {
	submission_node = readline(prompt = "What is the name of your submission node? ")
	user = readline(prompt = qq("What is your user name on @{submission_node}? "))

	bsub_opt$submission_node = submission_node
	bsub_opt$user = user

	# try to connect 
	ssh_disconnect()
	ssh_connect()

	LSF_ENVDIR = ssh_exec("echo $LSF_ENVDIR")
	LSF_SERVERDIR = ssh_exec("echo $LSF_SERVERDIR")

	bsub_opt$ssh_envir = c("source /etc/profile",
                           qq("export LSF_ENVDIR=@{LSF_ENVDIR}"),
                           qq("export LSF_SERVERDIR=@{LSF_SERVERDIR}"))

	R_version = readline(prompt = "What is the R version? ")
	bsub_opt$R_version = R_version

	call_Rscript = readline(prompt = "How is `Rscript` called, taking {version} as the variable for version? ")
	call_Rscript = gsub("\\{", "@{", call_Rscript)
	call_Rscript_fun = eval(parse(text = qq('function(version) qq("@{call_Rscript}")')))

	Rscript_call = call_Rscript_fun(paste0(R.version$major, ".", R.version$minor))
	oe = try(run_cmd(qq("@{Rscript_call} --version")), silent = TRUE)
	if(inherits(oe, "try-error")) {
		warning_wrap(qq("Cannot run `@{Rscript_call} --version` on the submission node, check your `bsub_opt$call_Rscript` global option."))
	}
	bsub_opt$call_Rscript = call_Rscript_fun

	if(verbose) {
		print(bconf)
	}
	
	invisible(NULL)
}


config_odcf = function(user = NULL, verbose = TRUE) {
	bsub_opt$call_Rscript = function(version) qq("module load gcc/7.2.0; module load java/1.8.0_131; module load R/@{version}; Rscript")
	bsub_opt$submission_node = c("bsub02.lsf.dkfz.de")
	bsub_opt$ssh_envir = c("source /etc/profile",
                           "export LSF_ENVDIR=/opt/lsf/conf",
                           "export LSF_SERVERDIR=/opt/lsf/10.1/linux3.10-glibc2.17-x86_64/etc")
	if(!is.null(user)) bsub_opt$user = user
	if(!file.exists("~/.bsub_temp")) {
		dir.create("~/.bsub_temp", recursive = TRUE, showWarnings = FALSE)
	}
	bsub_opt$temp_dir = "~/.bsub_temp"

	ssh_disconnect()

	if(verbose) qqcat("configure for user '@{bsub_opt$user}' on node @{paste(bsub_opt$submission_node, collapse = ', ')}.\n")
	invisible(NULL)
}


config_sanger_farm3_head3 = function(user = NULL, group = NULL, verbose = TRUE) {
	bsub_opt$call_Rscript = function(version) qq("Rscript")
	bsub_opt$submission_node = c("farm3-head3")
	bsub_opt$temp_dir = "~/.bsub_temp"

	# values for LSF_SERVERDIR and LSF_ENVDIR can be get by:
	# echo $LSF_SERVERDIR
	# echo $LSF_ENVDIR
	bsub_opt$ssh_envir = c("source /etc/profile",
                           "export LSF_ENVDIR=/usr/local/lsf/conf",
                           "export LSF_SERVERDIR=/usr/local/lsf/9.1/linux2.6-glibc2.3-x86_64/etc")
	if(!is.null(user)) bsub_opt$user = user
	if(is.null(group) && is.null(bsub_opt$group)) {
		bsub_opt$bsub_template = function(name, hours, memory, cores, output, ...) {
			qq("bsub -J '@{name}' -W '@{hours}:00' -n @{cores} -R 'select[mem>@{round(memory*1024)}] rusage[mem=@{round(memory*1024)}]' -M@{round(memory*1024)} -o '@{output}'")
		}
	} else {
		if(is.null(group) && !is.null(bsub_opt$group)) group = bsub_opt$group
		bsub_opt$bsub_template = function(name, hours, memory, cores, output, ...) {
			qq("bsub -J '@{name}' -W '@{hours}:00' -n @{cores} -R 'select[mem>@{round(memory*1024)}] rusage[mem=@{round(memory*1024)}]' -M@{round(memory*1024)} -G @{group} -o '@{output}'")
		}
	}

	ssh_disconnect()

	if(verbose) qqcat("configure for user '@{bsub_opt$user}' on node @{paste(bsub_opt$submission_node, collapse = ', ')}.\n")
	invisible(NULL)
}


config_sanger = function(user = NULL, ssh_key = "~/.ssh/id_rsa", verbose = TRUE) {
	bsub_opt$call_Rscript = function(version) qq("Rscript")
	bsub_opt$login_node = "ssh.sanger.ac.uk"
	bsub_opt$submission_node = NULL
	bsub_opt$temp_dir = "~/.bsub_temp"

	# values for LSF_SERVERDIR and LSF_ENVDIR can be get by:
	# echo $LSF_SERVERDIR
	# echo $LSF_ENVDIR
	if(!is.null(user)) bsub_opt$user = user

	ssh_envir = "source /etc/profile; export LSF_ENVDIR=/usr/local/lsf/conf;export LSF_SERVERDIR=/usr/local/lsf/9.1/linux2.6-glibc2.3-x86_64/etc"

	bsub_opt$ssh_envir = c(
		"source /etc/profile",
		qq("alias bjobs=\"ssh -i @{ssh_key} @{bsub_opt$user}@farm3-head3 '@{ssh_envir};bjobs'\""),
		qq("alias bparam=\"ssh -i @{ssh_key} @{bsub_opt$user}@farm3-head3 '@{ssh_envir};bparam'\""),
		qq("alias bkill=\"ssh -i @{ssh_key} @{bsub_opt$user}@farm3-head3 '@{ssh_envir};bkill'\""),
		qq("alias bsub=\"ssh -i @{ssh_key} @{bsub_opt$user}@farm3-head3 '@{ssh_envir};bsub'\"")
	)

	ssh_disconnect()

	if(verbose) qqcat("configure for user '@{bsub_opt$user}' on node @{paste(bsub_opt$login_node, collapse = ', ')}.\n")
	invisible(NULL)
}


config_two_ssh = function(user = NULL) {
	bsub_opt$call_Rscript = function(version) qq("module load gcc/7.2.0; module load java/1.8.0_131; module load R/@{version}; Rscript")
	bsub_opt$login_node = "odcf-worker02"
	bsub_opt$submission_node = NULL
	
	# values for LSF_SERVERDIR and LSF_ENVDIR can be get by:
	# echo $LSF_SERVERDIR
	# echo $LSF_ENVDIR
	if(!is.null(user)) bsub_opt$user = user

	ssh_envir = "source /etc/profile; export LSF_ENVDIR=/opt/lsf/conf;export LSF_SERVERDIR=/opt/lsf/10.1/linux3.10-glibc2.17-x86_64/etc"

	bsub_opt$ssh_envir = c(
		"source /etc/profile",
		qq("alias bjobs=\"ssh -i ~/.ssh/id_rsa_dkfz_heidelberg @{bsub_opt$user}@odcf-cn34u03s12 '@{ssh_envir};bjobs'\""),
		qq("alias bparam=\"ssh -i ~/.ssh/id_rsa_dkfz_heidelberg @{bsub_opt$user}@odcf-cn34u03s12 '@{ssh_envir};bparam'\""),
		qq("alias bkill=\"ssh -i ~/.ssh/id_rsa_dkfz_heidelberg @{bsub_opt$user}@odcf-cn34u03s12 '@{ssh_envir};bkill'\"")
	)

	ssh_disconnect()

	qqcat("configure for user '@{bsub_opt$user}' on node @{paste(bsub_opt$login_node, collapse = ', ')}.\n")
	invisible(NULL)
}

message_wrap = function (..., appendLF = TRUE) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    message(x, appendLF = appendLF)
}

warning_wrap = function (..., appendLF = TRUE) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    x = crayon::magenta(x)
    warning(x, call. = FALSE)
}

stop_wrap = function (...) {
    x = paste0(...)
    x = paste(strwrap(x), collapse = "\n")
    x = crayon::red(x)
    stop(x, call. = FALSE)
}


on_submission_node = function() {
    Sys.info()["nodename"] %in% bsub_opt$submission_node
}


under_same_file_system = function() {

    if(!is.null(ENV$on_same_file_system)) {
        return(ENV$on_same_file_system)
    }

    if(on_submission_node()) {
        ENV$on_same_file_system = TRUE
        return(TRUE)
    }

    f = tempfile(tmpdir = normalizePath("~"))
    f = paste0(f, sample(10000000, 1))
    file.create(f)
    on.exit(file.remove(f))

    # now check the submission node
    oe = try(ln <- run_cmd(qq("ls @{f}")), silent = TRUE)

    if(inherits(oe, "try-error")) {
        ENV$on_same_file_system = FALSE
        return(FALSE)
    } else {
        ENV$on_same_file_system = TRUE
        return(TRUE)
    }
}

#' Rerun jobs
#' 
#' @param job_id A single job ID. In `pipeline_rerun()`, it is the job ID of any one job in the pipeline.
#' @param dependency A vector of job IDs that current job depends on.
#' @param verbose Whether to print messages.
#' @param job_tb The data frame returned from [`bjobs()`], internally used.
#' 
#' @returns
#' `job_rerun()` returns the job IDs. `pipeline_rerun()` returns `NULL`.
#' @rdname rerun
#' @export
job_rerun = function(job_id, dependency = character(0), verbose = TRUE, job_tb = NULL) {

	if(is.null(job_tb)) {
		job_tb = bjobs(status = "all", print = FALSE, job_id = job_id)
	} else {
		job_tb = job_tb[job_tb$JOBID == job_id, , drop = FALSE]
	}

	if(verbose) message(qq("rerun job @{job_id} <@{job_tb$JOB_NAME}>"))
	if(verbose) message("  - obtain submission script")

	log = run_cmd(qq("bjobs -script @{job_id}"))
    i1 = grep("^# LSBATCH: User input", log)
    source = FALSE
    if(log[i1+1] == "( cat <<_USER_\\SCRIPT_") {
        i1 = i1 + 1
        source = TRUE
    }
    if(source) {
        i2 = which(log == "_USER_SCRIPT_")
        script = log[seq(i1+1, i2-1)]
    } else {
        i2 = grep("^# LSBATCH: End user input", log)
        script = log[seq(i1+1, i2-1-2)]
    }

    if(source) {
    	i1 = which(log == "##### commands start #####")
    	if(length(i1)) {
    		i2 = which(log == "##### commands end #####")
    		script = log[seq(i1+1, i2-1)]
    	}
    }

    if(verbose) message("  - obtain environment variables")
    ev = job_env(job_id)
    # ev = ev[names(ev) %in% env_var]
	
	if(verbose) message("  - obtain resource requirement")

    name = job_tb$JOB_NAME
    mem = as.numeric(gsub("^.*mem=(.*?)\\].*$", "\\1", job_tb$COMBINED_RESREQ))
    cores = job_tb$NREQ_SLOT
    walltime = job_tb$RUNTIMELIMIT
    walltime = ceiling(walltime)

    mem = ceiling(mem/1024)
    hour = ceiling(walltime/60)
     
    output_dir = dirname(job_output_file_by_id(job_id))
    gl = job_attached_vars_by_id(job_id)
    if(is.na(gl$temp_dir)) {
    	temp_dir = output_dir
    } else {
    	temp_dir = gl$temp_dir
    }

    if(under_same_file_system()) {

    	if(verbose) message("  - save into a temporary bash file")
    	sh_file = tempfile(fileext = ".sh")
    
	    con = file(sh_file, "w")
	    writeLines(paste0("export ", ev), con)
	    writeLines("\n", con)
	    writeLines(script, con)
	    close(con)

	    if(verbose) message("  - directly submit the new job")
		job_id = bsub_cmd(sh = sh_file, name = name, memory = mem, hours = hour, cores = cores, dependency = dependency, ask = FALSE)
		file.remove(sh_file)

    } else {
    	if(verbose) message("  - save into a temporary bash file")
    	t = as.POSIXlt(Sys.time())
	    t = as.numeric(t) + t$sec - floor(t$sec)
	    t = gsub("\\.", "_", t)
	    name_uid = paste0(name, "_", t)
	    output = qq("@{output_dir}/@{name_uid}.out")

    	# upload to submission node
    	local_sh_file = tempfile(paste0(name_uid, "_"), tmpdir = tempdir(), fileext = ".sh")
    	con = file(local_sh_file, "w")
	    writeLines(paste0("export ", ev), con)
	    writeLines("\n", con)
	    writeLines(script, con)
	    close(con)

	    if(verbose) message("  - upload to the submission node")
	    sh_file = paste0(temp_dir, "/", basename(local_sh_file))
    	ssh::scp_upload(bsub_opt$ssh_session, local_sh_file, to = sh_file, verbose = FALSE)
    	file.remove(local_sh_file)

    	if(verbose) message("  - submit the new job")
    	cmd = bsub_opt$bsub_template(name, hour, mem, cores, output, bsub_opt$group)
    	if(length(dependency)) {
	        dependency_str = paste( paste("done(", dependency, ")"), collapse = " && " )
	        cmd = qq("@{cmd} -w '@{dependency_str}' \\\n")
	    }
    	cat(silver(qq("@{cmd} < '@{sh_file}'")), "\n")

	    # add values of name_uid and temp_dir to `bjobs` in job description
	    cmd = qq("@{cmd} -env 'all,R_BSUB_NAME_UID=@{name_uid},R_BSUB_TEMP_DIR=@{temp_dir}' -Jd 'R_BSUB_NAME_UID=@{name_uid},R_BSUB_TEMP_DIR=@{temp_dir}'")
	    cmd = qq("@{cmd} < '@{sh_file}'; rm @{sh_file}")
	    txt = run_cmd(cmd, print = FALSE)

	    job_id = gsub("^.*<(\\d+)>.*$", "\\1", txt)

    }
    
    job_id
}

job_env = function(job_id) {
	ev = run_cmd(qq("bjobs -env @{job_id}"))
	ev = ev[!grepl("\\(\\)=", ev)]
	ev = ev[!grepl("^}", ev)]
	ev = ev[!grepl("=$", ev)]
	ev = ev[!grepl(" ", ev)]
	ev = ev[!grepl("R_BSUB_NAME_UID", ev)]
	ev = ev[!grepl("R_BSUB_TEMP_DIR", ev)]
	ev = ev[!grepl("LS_COLORS", ev)]

	l = grepl(";", ev)
	if(any(l)) {
		ev[l] = gsub("^(\\S+)=(.*)$", "\\1=\"\\2\"", ev[l])
	}

	env_var_name = gsub("^(\\S+)=.*$", "\\1", ev)
	names(ev) = env_var_name
	ev
}

#' @param skip_done Whether to skip done jobs.
#' 
#' @details
#' In `pipeline_rerun()`, the full set of jobs can be captured by one job in the pipeline.
#' 
#' @rdname rerun
#' @export
pipeline_rerun = function(job_id, skip_done = TRUE, verbose = TRUE) {

	if(verbose) message("rerun pipeline")
	job_tb = bjobs(status = "all", print = FALSE)
	g = job_dependency_igraph(job_id, job_tb)
	all_job_ids = V(g)$name

	job_tb2 = job_tb[job_tb$JOBID %in% all_job_ids, , drop = FALSE]
	rownames(job_tb2) = job_tb2$JOBID
	job_tb2 = job_tb2[order(as.numeric(job_tb2$JOBID)), , drop = FALSE]

	l_run_pend = job_tb2$STAT %in% c("RUN", "PEND")
	if(any(l_run_pend)) {
		message("Found following RUN/PEND jobs:")
		for(i in which(l_run_pend)) {
			message(qq("  job @{job_tb2$JOBID[i]} <@{job_tb2$JOB_NAME[i]}>, @{job_tb2$STAT[i]}"))
		}
		stop_wrap("There should be no RUN/PEND job if rerun the pipeine.")
	}

	if(length(all_job_ids) == 1) {
		if(job_tb2$STAT[1] == "DONE") {
			if(skip_done) {
				message(qq("skip DONE job @{job_tb2$JOBID[1]} <@{job_tb2$JOB_NAME[1]}>"))
			} else {
				job_rerun(job_id, verbose = verbose, job_tb = job_tb)
			}
		} else {
			job_rerun(job_id, verbose = verbose, job_tb = job_tb)
		}
		return(invisible(NULL))
	}

	nr = nrow(job_tb2)
	done_jobs = structure(rep(FALSE, nr), names = job_tb2$JOBID)
	done_jobs[job_tb2$STAT == "DONE"] = TRUE

	new_job_id = character(0)
	for(i in seq_len(nr)) {
		if(job_tb2$STAT[i] == "DONE" && skip_done) {
			# skip
			if(verbose) {
				message(qq("skip DONE job @{job_tb2$JOBID[i]} <@{job_tb2$JOB_NAME[i]}>"))
			}
		} else {
			d = distances(g, which(V(g)$name == job_tb2$JOBID[i]), to = V(g), mode = "in")
			p = colnames(d)[which(d == 1)]
			if(skip_done) {
				dj = done_jobs[p]
				dj = dj[dj]
				p = setdiff(p, names(dj))
			}
			dep = character(0)
			if(length(p)) {
				dep = new_job_id[p]
			}
			jid = job_rerun(job_tb2$JOBID[i], dependency = dep, verbose = verbose, job_tb = job_tb)
			new_job_id[as.character(job_tb2$JOBID[i])] = jid
		}
	}

	return(invisible(NULL))
}

#' Kill jobs
#'
#' @param job_id A vector of job IDs or a data frame returned by [`bjobs()`].
#' @param filter Regular expression on job names (only the running and pending jobs). It is
#'      only used when `job_id` is not set.
#'
#' @export
#' @examples
#' \dontrun{
#' job_id = c(10000000, 10000001, 10000002)  # job ids can be get from `bjobs()`
#' bkill(job_id)
#' # kill all jobs (running and pending) of which the names contain "example"
#' bkill(filter = "example") 
#' }
bkill = function(job_id, filter = NULL) {

    if(missing(job_id)) {
        job_df = bjobs(status = c("RUN", "PEND"), max = Inf, filter = filter, print = FALSE)
        job_id = job_df$JOBID
    }

    if(is.data.frame(job_id)) {
        df = job_id
        df = df[df$STAT %in% c("RUN", "PEND"), , drop = FALSE]
        if(nrow(df)) {
            job_id = df$JOBID
        } else {
            return(invisible(NULL))
        }
    }

    cmd = qq("bkill @{paste(job_id, collapse = ' ')} 2>&1")

    run_cmd(cmd, print = TRUE)
}

#' Run command on submission node
#'
#' @param cmd A single-line command.
#' @param print Whether to print output from the command.
#'
#' @details
#' If current node is not the submission node, the command is executed via ssh.
#'
#' @returns The output of the command.
#' @export
#' @examples
#' \dontrun{
#' # run pwd on remote node
#' run_cmd("pwd")
#' }
run_cmd = function(cmd, print = FALSE) {
    if(on_submission_node()) {
       con = pipe(cmd)
       ln = readLines(con)
       close(con)
    } else {
        ln = ssh_exec(cmd)
    }

    if(print) cat(ln, sep = "\n")
    return(invisible(ln))
}

 



#' Parameters for bsub
#'
#' @param ... Arguments for the parameters, see "details" section
#' @param RESET reset to default values
#' @param READ.ONLY please ignore
#' @param LOCAL please ignore
#' @param ADD please ignore
#' 
#' @details
#' There are following global parameters:
#' 
#' - `packages`: A character vector with package names that will be loaded before running the script.
#' - `image`: A character vector of RData/rda files that will be loaded before running the script.
#' - `temp_dir`: Path of temporary folder where the temporary R/bash scripts will be put.
#' - `output_dir`: Path of output folder where the output/flag files will be put.
#' - `enforce`: If a flag file for the job is found, whether to enforce to rerun the job.
#' - `R_version`: The version of R.
#' - `working_dir`: The working directory.
#' - `ignore`: Whether ignore `bsub_chunk`, `bsub_script` and `bsub_cmd`.
#' - `local`: Run job locally (not submitting to the LSF cluster)?
#' - `call_Rscript`: How to call `Rscript` by specifying an R version number.
#' - `submission_node`: A list of node names for submitting jobs.
#' - `login_node`: This value basically is the same as ``submission_node`` unless the login nodes are different from submission nodes.
#' - `sh_head`: Commands that are written as head of the sh script.
#' - `user`: Username on the submission node.
#' - `group`: The user group
#' - `ssh_envir`: The commands for setting bash environment for successfully running bjobs, bsub, ...
#' - `bsub_template`: Template for constructing `bsub` command.
#' - `parse_time`: A function that parses time string from the LSF `bjobs` command to a `POSIXct` object.
#' - `verbose`: Whether to print more messages.
#'
#' `ssh_envir` should be properly set so that LSF binaries such as `bsub` or `bjobs` can be properly found.
#' There are some environment variables initialized when logging in the bash terminal while they are not initialized with the
#' ssh connection. Thus, some environment variables should be manually set.
#'
#' An example for `ssh_envir` is as follows. The `LSF_ENVDIR` and `LSF_SERVERDIR` should be defined and exported.
#' 
#' ```r
#' c("source /etc/profile",
#'   "export LSF_ENVDIR=/opt/lsf/conf",
#'   "export LSF_SERVERDIR=/opt/lsf/10.1/linux3.10-glibc2.17-x86_64/etc")
#'```
#' 
#' The values of these two variables can be obtained by entering following commands in your bash terminal (on the submission node):
#'
#' ```
#' echo $LSF_ENVDIR
#' echo $LSF_SERVERDIR
#'```
#' 
#' The time strings by LSF `bjobs` command might be different for different configurations. The **bsub**
#' package needs to convert the time strings to `POSIXlt` objects for calculating the time difference. Thus, if
#' the default time string parsing fails, users need to provide a user-defined function and set with ``parse_time``
#' option in `bsub_opt`. The function accepts a vector of time strings and returns a `POSIXlt` object. For example,
#' if the time string returned from ``bjobs`` command is in a form of ``Dec 1 18:00:00 2019``, the parsing function
#' can be defined as:
#'
#' ```r
#' bsub_opt$parse_time = function(x) {
#'     as.POSIXlt(x, format = "\\%b \\%d \\%H:\\%M:\\%S \\%Y")
#' }
#' ```
#' 
#' @returns The corresponding option values.
#' @import GlobalOptions
#' @import GetoptLong
#' @export
#' @examples
#' # The default bsub_opt
#' bsub_opt
bsub_opt = set_opt(
    ask = list(
        .value = TRUE,
        .visible = FALSE
    ),
    packages = list(
        .class = "character"
    ),
    image = list(
        .class = "character",
        .filter = function(x) {
            l = file.exists(x)
            if(all(!l)) {
                message_wrap(qq("All the image files specified do not exist. Save current workspace to @{x[1]}? [y/n]"), appendLF = FALSE)
                answer = readline()
                if(answer %in% c("y", "Y", "yes", "Yes")) {
                    save.image(file = x[1])
                    x = x[1]
                    l = TRUE
                } else {
                    stop("All image files do not exist.")
                }

            }
            if(any(!l)) {
                stop(qq("Some of the image files do not exist."))
            }
            normalizePath(x)
        }
    ),
    temp_dir = list(
        .value = "~/.bsub_temp",
        .validate = function(x) {
            check_temp_dir(x, ask = .v$ask)
            TRUE
        },
        .filter = function(x) {
            normalizePath(x)
        }
    ),
    output_dir = list(
        .value = function() .v$temp_dir,
        .validate = function(x) {
            if(!file.exists(x) && .v$ask) {
                answer = readline(qq("create output_dir: @{x}? [y|n] "))
                if(answer %in% c("y", "Y", "yes", "Yes", "YES")) {
                    dir.create(x, recursive = TRUE, showWarnings = FALSE)
                } else {
                    stop_wrap(qq("not allowed to create @{x}."))
                }
            }
            TRUE
        },
        .filter = function(x) {
            normalizePath(x)
        }
    ),
    enforce = list(
        .value = TRUE,
        .class = "logical"
    ),
    R_version = list(
        .value = qq("@{R.version$major}.@{R.version$minor}")
    ),
    working_dir = list(
        .value = "",
        .validate = file.exists,
        .length = 1
    ),
    wd = list(.synonymous = "working_dir"),
    ignore = list(
        .value = FALSE,
        .class = "logical"
    ),
    local = list(
        .value = FALSE,
        .class = "logical"
    ),
    call_Rscript = list(
        .value = function(version) "Rscript",
        .class = "function"
    ),
    submission_node = list(
        .value = NULL
    ),
    login_node = list(
        .value = function() .v$submission_node
    ),
    ssh_session = list(
        .value = NULL,
        .private = TRUE,
        .visible = FALSE
    ),
    debug = list(
        .value = TRUE,
        .visible = FALSE
    ),
    sh_head = list(
        .value = "",
        .class = "character"
    ),
    user = list(
        .value = Sys.info()['user']
    ),
    group = list(
        .value = NULL
    ),
    ssh_envir = list(
    	.value = c("source /etc/profile")
    ),
    bsub_template = list(
        .value = function(name, hour, memory, core, output, group = NULL, ...) {
            if(identical(group, "") || identical(group, NULL)) {
                cmd = qq("bsub -J '@{name}' -W '@{floor(hour)}:@{floor((hour-floor(hour))*60)}' -n @{core} -R 'rusage[mem=@{round(memory*1024)}]'", 
                         "     -o '@{output}'", 
                         "   ", sep = " \\\n")
            } else {
                cmd = qq("bsub -J '@{name}' -W '@{floor(hour)}:@{floor((hour-floor(hour))*60)}' -n @{core} -R 'rusage[mem=@{round(memory*1024)}]'",
                         "     -G @{bsub_opt$group} -o '@{output}'",
                         "   ", sep = " \\\n")
            }
            return(cmd)
        },
        .class = "function"
    ),
    parse_time = list(
        .value = NULL,
        .class = "function"
    ),
    verbose = list(
        .value = FALSE,
        .class = "logical"
    ),
    history_timestamp = list(
        .value = NULL,
        .visible = FALSE
    )
)

check_temp_dir = function(x, ask = TRUE) {
    if(!file.exists(x) && ask) {
        answer = readline(qq("create temp_dir: @{x}? [y|n] "))
        if(answer %in% c("y", "Y", "yes", "Yes", "YES")) {
            dir.create(x, recursive = TRUE, showWarnings = FALSE)
        } else {
            stop_wrap(qq("not allowed to create @{x}."))
        }
    } else {
        all_f = list.files(x, full.names = TRUE)

        if(length(all_f)) {
            file_size = sum(file.info(all_f)[, "size"], na.rm = TRUE)
            if(file_size < 1024) {
                fs = paste0(file_size, "Byte")
            } else if(file_size < 1024^2) {
                fs = paste0(round(file_size/1024, 1), "KB")
            } else if(file_size < 1024^3) {
                fs = paste0(round(file_size/1024^2, 1), "MB")
            } else {
                fs = paste0(round(file_size/1024^3, 1), "GB")
            }
            message(qq("There are @{length(all_f)} temporary files (@{fs}) in @{x}"))
        }
    }
}

#' Print current configurations
#'
#' @details
#' This function is only for printing. Use [`bsub_opt()`] to change configurations.
#'
#' You simply type `bconf` (without the brackets) in the interactive R console.
#'
#' @returns A `bconf` object.
#'
#' @export
#' @rdname bconf
#' @docType data
#' @examples
#' bconf
bconf = structure(NA, class = "bconf")

#' @param x A `bconf` object.
#' @param ... Other parameters.
#' 
#' @export
#' @rdname bconf
print.bconf = function(x, ...) {
    cat(get_bconf_message(x), "\n")

    flag = 0
    msg = NULL
    if(!file.exists(bsub_opt$temp_dir)) {
        msg = c(msg, crayon::red(qq("!! The temp_dir (@{bsub_opt$temp_dir}) does not exist.")))
        flag = 1
    }
    if(bsub_opt$temp_dir != bsub_opt$output_dir) {
        if(!file.exists(bsub_opt$output_dir)) {
            msg = c(msg, crayon::red(qq("!! The output_dir (@{bsub_opt$output_dir}) does not exist.")))
            flag = 2
        }
    }
    if(flag) {
        if(flag == 1) {
            msg = c(msg, crayon::red("!! The directory can be created manually, or by setting a value to"),
                         crayon::red("!! `bsub_opt$temp_dir`, or by the first call of `bsub_*()`."))
        } else {
            msg = c(msg, crayon::red("!! The directory can be created manually, or by setting a value to"),
                         crayon::red("!! `bsub_opt$output_dir`, or by the first call of `bsub_*()`."))
        }

        cat("\n")
        cat(msg, sep = "\n")
        cat("\n")
    }
}

get_bconf_message = function(x, ...) {
    x = bsub_opt()
    msg = NULL
    msg = c(msg, "Configurations for bsub:")
    if(is.null(x$user)) {
        msg = c(msg, "  * user is not defined")
    } else {
        msg = c(msg, qq("  * user for connecting submission node: @{crayon::green(x$user)}"))
    }

    if(!is.null(x$group)) {
        msg = c(msg, qq("  * user group : @{crayon::green(x$group)}"))
    }

    if(is.null(x$submission_node)) {
        msg = c(msg, qq(crayon::red("  ! submission node is not defined")))
    } else {
        msg = c(msg, qq("  * submission node: @{crayon::green(paste(x$submission_node, collapse = ', '))}"))

        if(length(setdiff(x$login_node, x$submission_node))) {
            msg = c(msg, qq("  * login node: @{crayon::green(paste(x$login_node, collapse = ', '))}"))
        }
    }
    
    if(!is.null(x$R_version)) {
        msg = c(msg, qq("  * global R version: @{crayon::green(x$R_version)}"))
    }

    msg = c(msg, "  * command to call `Rscript`:")
    msg = c(msg, crayon::green(paste0("     ", deparse(body(x$call_Rscript)), " foo.R")))

    if(!is.null(x$package)) {
        msg = c(msg, qq("  * Global R packages: @{crayon::green(paste(x$packages, collapse=','))}"))
    }

    if(!is.null(x$image)) {
        msg = c(msg, "  * Global image files that will be loaded to every job:")
        for(f in x$image) {
            msg = c(msg, crayon::green(qq("    - @{f}")))
        }
    }

    msg = c(msg, qq("  * temporary directory: @{crayon::green(x$temp_dir)}"))

    if(!x$enforce) {
        msg = c(msg, "  - successful jobs are skipped.")
    }

    msg = c(msg, "")
    msg = c(msg, "Configurations can be modified by `bsub_opt()` function")
    return(paste(msg, collapse = "\n"))
}


STATUS_COL = c("RUN" = "darkgreen", "EXIT" = "#e41a1c", "PEND" = "#ff7f00", "DONE" = "#377eb8", "unknown" = "grey")
ENV = new.env()
ENV$on_same_file_system = NULL

#' Submit jobs
#'
#' @param code The code chunk, it should be embraced by `{}`.
#' @param name If name is not specified, an internal name calculated by [`digest::digest()`] on the chunk is automatically assigned. 
#' @param packages A character vector with package names that will be loaded before running the script. There is a special name `_in_session_`
#'     that loads all the packages loaded in current R session.
#' @param image A character vector of `.RData`/`.rda` files that will be loaded before running the script.
#'       When `image` is set to `TRUE`, all variables in [`.GlobalEnv`] will be saved
#'       into a temporary file and all attached packages will be recorded. The temporary files
#'       will be removed after the job is finished.
#' @param variables A character vector of variable names that will be loaded before running the script. There is a special name `_all_functions_`
#'       that saves all functions defined in the global environment.
#' @param share A character vector of variables names for which the variables are shared between jobs. Note the temporary `.RData` files are not deleted
#'        automatically.
#' @param working_dir The working directory.
#' @param hours Running time of the job.
#' @param memory Memory usage of the job. It is measured in GB.
#' @param cores Number of cores.
#' @param R_version R version.
#' @param temp_dir Path of temporary folder where the temporary R/bash scripts will be put.
#' @param output_dir Path of output folder where the output/flag files will be put.
#' @param dependency A vector of job IDs that current job depends on.
#' @param enforce If a flag file for the job is found, whether to enforce to rerun the job.
#' @param local Run job locally (not submitting to the LSF cluster)?
#' @param script In `bsub_chunk()`, it is the path of a script where code chunks will be extracted and sent to the cluster. 
#'       It is always used with `start` and `end` arguments. In `bsub_script()`, it is the path of the R script to submit.
#' @param start A numeric vector that contains line indices of the starting code chunk or a character vector
#'        that contain regular expression to match the start of code chunks.
#' @param end Same setting as `start`.
#' @param save_var Whether save the last variable in the code chunk? Later the variable
#'    can be retrieved by [`retrieve_var()`].
#' @param sh_head Commands that are written as head of the sh script.
#' @param ask Whether to promote.
#' 
#' @details
#' `job_chunk()` submits R code chunk.
#' 
#' `job_script()` submits R script with command-line arguments.
#' 
#' `job_cmd()` submits general bash commands.
#'
#' @returns A job ID.
#' @export
#' @importFrom codetools findGlobals
#' @importFrom utils sessionInfo find
#' @import crayon
#' @rdname bsub
#' @examples
#' \dontrun{
#' bsub_chunk(name = "example", memory = 10, hours = 10, cores = 4,
#' {
#'     Sys.sleep(5)
#' })
#' 
#' # the R version is defined in bsub_opt$R_version
#' bsub_script("/path/of/foo.R", name = ..., memory = ..., cores = ..., ...)
#' # with command-line arguments
#' bsub_script("/path/of/foo.R", argv = "--a 1 --b 3", ...)
#' 
#' # put all arguments also in the command
#' bsub_cmd("some-tool -arg1 1 -arg2 2", name = ..., memory = ..., cores = ..., ...)
#' }
bsub_chunk = function(code, 
    name = NULL,
    packages = bsub_opt$packages, 
    image = bsub_opt$image,
    variables = character(),
    share = character(),
    working_dir = bsub_opt$working_dir,
    hours = 1, 
    memory = 1, 
    cores = 1,
    R_version = bsub_opt$R_version,
    temp_dir = bsub_opt$temp_dir,
    output_dir = bsub_opt$output_dir,
    dependency = NULL,
    enforce = bsub_opt$enforce,
    local = bsub_opt$local,
    script = NULL,
    start = NULL,
    end = NULL,
    save_var = FALSE,
    sh_head = bsub_opt$sh_head,
    ask = TRUE) {

    if(!under_same_file_system()) {
        stop("Job can only be sumitted on the same file system as submission nodes.")
    }
    
    if(bsub_opt$ignore) return(invisible(NULL))

    if(!is.null(script)) {
        script = readLines(script)
        if(is.null(start)) start = 1
        if(is.null(end)) end = length(script)
        if(is.character(start)) {
            start = grep(start, script)
            if(length(start) == 0) {
                stop(qq("cannot find start mark '@{start}'"))
            }
            if(length(start) > 1) {
                stop(qq("find more than one start marks '@{start}'"))
            }
        }
        if(is.character(end)) {
            end = grep(end, script)
            if(length(end) == 0) {
                stop(qq("cannot find end mark '@{end}'"))
            }
            if(length(end) > 1) {
                stop(qq("find more than one end marks '@{end}'"))
            }
        }
        if(end < start) {
            stop(qq("start line (#@{start}) should be no larger than the end line (#@{end})."))
        }
        code = script[start:end]
        code_validated = FALSE
    } else {

        if(length(variables) == 0) {
            expr = substitute(code)
            code_fun = function() {}
            body(code_fun) = expr
            gl = findGlobals(code_fun, merge = FALSE)
            variables = gl$variables

            if(length(variables)) {
                message_wrap(qq("There are variables (@{paste(variables, collapse = ', ')}) not defined in the code chunk. Automatically use the ones from global environment. Note this is approximate."))
            }

            if(length(packages) == 0 && !identical(image, TRUE)) {
                functions = gl$functions
                for(f in functions) {
                    p = find(f)
                    if(length(p)) {
                        if(grepl("package:", p[1])) {
                            packages = c(packages, p[1])
                        } else if(p[1] == ".GlobalEnv") {
                            variables = c(variables, f)
                        }
                    }
                }
                packages = gsub("package:", "", packages)
                packages = unique(packages)
                packages = setdiff(packages, c("base", "stats", "graphics", "grDevices", "utils", "datasets", "methods"))
                if(length(packages)) {
                    message_wrap(qq("There are packages (@{paste(packages, collapse = ', ')}) not specified in the code chunk. Automatically detech them. Note this is approximate."))
                }
            }

            code_validated = TRUE

            code = deparse(expr)
        } else {
            code = deparse(substitute(code))
            code_validated = FALSE
        }
    }
    code = paste(code, collapse = "\n")
    code = paste0(code, "\n")

    if(!code_validated) {
        expr = parse(text = code)
        code_fun = function() {}
        body(code_fun) = expr
        gl = findGlobals(code_fun, merge = FALSE)
        variables_unimported = setdiff(gl$variables, variables)

        if(length(variables_unimported)) {
            warning_wrap(qq("There are variables (@{paste(variables_unimported, collapse = ', ')}) not defined and not imported in the code chunk."))
        }

        p2 = character(0)
        if(!identical(image, TRUE)) {
            functions = gl$functions
            for(f in functions) {
                p = find(f)
                if(length(p)) {
                    if(grepl("package:", p[1])) {
                        p2 = c(p2, p[1])
                    } else if(p[1] == ".GlobalEnv") {
                        variables = c(variables, f)
                    }
                }
            }
            packages = union(gsub("package:", "", p2), packages)
            packages = setdiff(packages, c("base", "stats", "graphics", "grDevices", "utils", "datasets", "methods"))
            if(length(packages)) {
                warning_wrap(qq("There are packages (@{paste(packages, collapse = ', ')}) not specified in the code chunk."))
            }
        }
    }

    if(is.null(name)) {
        name = paste0("R_code_", digest::digest(code, "crc32"))
    } else {
        name = qq(name)
    }

    t = as.POSIXlt(Sys.time())
    t = as.numeric(t) + t$sec - floor(t$sec)
    t = gsub("\\.", "_", t)
    name_uid = paste0(name, "_", t)

    if(!file.exists(temp_dir)) {
        if(ask) {
            answer = readline(qq("create temp_dir: @{temp_dir}? [y|n] "))
            if(answer %in% c("y", "Y", "yes", "Yes", "YES")) {
                dir.create(temp_dir, recursive = TRUE, showWarnings = FALSE)
            } else {
                stop_wrap(qq("not allowed to create @{temp_dir}."))
            }
        } else {
            message("create temp_dir: ", temp_dir)
            dir.create(temp_dir, recursive = TRUE, showWarnings = FALSE)
        }
    }
    if(!file.exists(output_dir)) {
        if(ask) {
            answer = readline(qq("create output_dir: @{output_dir}? [y|n] "))
            if(answer %in% c("y", "Y", "yes", "Yes", "YES")) {
                dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
            } else {
                stop_wrap(qq("not allowed to create @{output_dir}."))
            }
        } else {
            message("create output_dir: ", output_dir)
            dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
        }
    }

    temp_dir = normalizePath(temp_dir)
    output_dir = normalizePath(output_dir)

    cat(cyan(qq("- job: '@{magenta(name)}' from a code chunk\n")))

    if(length(variables)) {
        if(!inherits(variables, "character")) {
            stop_wrap("`variables` must be a character vector (the names of the variables you want to import).")
        }
    }

    if(length(share)) {
        if(!inherits(share, "character")) {
            stop_wrap("`share` must be a character vector (the names of the variables you want to import).")
        }
    }

    # if `image` is true, save all variables and all packages that are loaded
    if(identical(image, TRUE)) {
        variables = c(variables, ls(envir = parent.frame(), all.names = TRUE))
        packages = c(rev(names(sessionInfo()$otherPkgs)), packages)
        image = NULL
    }

    ## a special tag for variables
    if(length(variables)) {
        if(any(grepl("^_all_functions", variables))) {
            var_name = ls(envir = parent.frame(), all.names = TRUE)
            l = sapply(var_name, function(x) inherits(get(x, envir = parent.frame()), "function"))
            var_name = var_name[l]
            variables = c(variables, var_name)
        }
    }

    ## a special tag for package: _in_session_
    if(length(packages)) {
        if(any(grepl("^_in_session", packages))) {
            packages = c(packages, rev(names(sessionInfo()$otherPkgs)))
        }
    }

    # analyze code

    head = "############## temporary R script ##############"
    tail = ""
    for(p in unique(packages)) {
        head = c(head, qq("library(@{p})"))
    }

    # if all image files do not exist, write current session to image[1]
    if(!is.null(image)) {
        l = file.exists(image)
        if(all(!l)) {
            message_wrap(qq("All the image files specified do not exist, save current workspace to @{image[1]}? [y/n]"), appendLF = FALSE)
            answer = readline()
            if(answer %in% c("y", "Y", "yes", "Yes")) {
                save.image(file = image[1])
                image = image[1]
            } else {
                stop("All image files do not exist.")
            }
        }

        for(f in image) {
            f = normalizePath(f)
            head = c(head, qq("load(\"@{f}\")"))
        }
    }

    if(length(variables)) {
        save(list = unique(variables), envir = parent.frame(), file = qq("@{temp_dir}/@{name_uid}_var.RData"))
        head = c(head, qq("load(\"@{temp_dir}/@{name_uid}_var.RData\")"))
        tail = c(tail, qq("invisible(file.remove(\"@{temp_dir}/@{name_uid}_var.RData\"))"))
    }

    if(length(share)) {
        for(i in seq_along(share)) {
            share_hash = digest::digest(list(name = share[i], value = get(share[i], envir = parent.frame())))
            share_file = qq("@{temp_dir}/@{name_uid}_var_shared_@{share_hash}.RData")
            if(!file.exists(share_file)) {
                save(list = share[i], envir = parent.frame(), file = share_file)
            }
            head = c(head, qq("load(\"@{share_file}\")"))
        }
    }

    tmp = tempfile(paste0(name_uid, "_"), fileext = ".R", tmpdir = temp_dir)
    if(bsub_opt$debug) {  # print R code to out file
        head = c(head, qq("cat(readLines(\"@{tmp}\"), sep = \"\\n\")"))
    }
    tail = c(tail, qq("invisible(file.remove(\"@{tmp}\"))\n\n"))  # the R script is already loaded in memory, the temp R script can be deleted

    if(save_var) {
        code = paste0("R_BSUB_RETURN_VAR = ", code)
        tail = c(tail, qq("saveRDS(R_BSUB_RETURN_VAR, file = '@{output_dir}/@{name_uid}_returned_var.rds')"))
    }

    tail = c(tail, "invisible(NULL)\n")

    if(working_dir != "") {
        if(!file.exists(working_dir)) {
            stop(qq("'@{working_dir}'' does not exist."))
        }
        head = c(head, qq("setwd('@{working_dir}')\n\n"))
    }
    writeLines(c(head, 
                 "#### R chunk start ####",
                 code, 
                 "#### R chunk end ####",
                 tail), con = tmp)

    command = qq("@{bsub_opt$call_Rscript(R_version)} '@{tmp}';")
    
    bsub_submit(
        command = command, 
        hours = hours,
        memory = memory,
        cores = cores,
        name = name,
        name_uid = name_uid,
        output_dir = output_dir,
        temp_dir = temp_dir,
        dependency = dependency,
        enforce = enforce,
        local = local,
        sh_head = sh_head)
}

#' @param argv A string of command-line arguments.
#' @param ... Command-line arguments can also be specified as name-value pairs.
#' 
#' @rdname bsub
#' @export
bsub_script = function(script, 
    argv = "", 
    name = NULL, 
    working_dir = bsub_opt$working_dir,
    hours = 1, 
    memory = 1, 
    cores = 1,
    R_version = bsub_opt$R_version,
    temp_dir = bsub_opt$temp_dir,
    output_dir = bsub_opt$output_dir,
    dependency = NULL,
    enforce = bsub_opt$enforce, 
    local = bsub_opt$local,
    sh_head = bsub_opt$sh_head,
    ask = TRUE,
    ...) {

    if(!under_same_file_system()) {
        stop("Job can only be sumitted on the same file system as submission nodes.")
    }
    
    if(bsub_opt$ignore) return(invisible(NULL))

    script = normalizePath(script)
    
    argv = qq(argv)

    argv_list = list(...)
    if(length(argv_list)) {
        argv_nm = names(argv_list)
        if(!all(grepl("^\\.", argv_nm))) {
            stop("Other command-line parameters specified in ... should have prefix '.'.")
        }

        argv_list = sapply(argv_list, function(x) {
                x = paste(x, collapse = " ")
        })

        argv_nm = gsub("^\\.", "", argv_nm)
        argv2 = qq("--@{argv_nm} @{argv_list} ")
        if(argv == "") {
            argv = argv2
        } else {
            argv = paste0(argv, " ", argv2)
        }
    }

    if(is.null(name)) {
        name = paste0("R_script_", digest::digest(list(script, argv), "crc32"))
    } else {
        name = qq(name)
    }

    t = as.POSIXlt(Sys.time())
    t = as.numeric(t) + t$sec - floor(t$sec)
    t = gsub("\\.", "_", t)
    name_uid = paste0(name, "_", t)

    if(!file.exists(temp_dir)) {
        if(ask) {
            answer = readline(qq("create temp_dir: @{temp_dir}? [y|n] "))
            if(answer %in% c("y", "Y", "yes", "Yes", "YES")) {
                dir.create(temp_dir, recursive = TRUE, showWarnings = FALSE)
            } else {
                stop_wrap(qq("not allowed to create @{temp_dir}."))
            }
        } else {
            message("create temp_dir: ", temp_dir)
            dir.create(temp_dir, recursive = TRUE, showWarnings = FALSE)
        }
    }
    if(!file.exists(output_dir)) {
        if(ask) {
            answer = readline(qq("create output_dir: @{output_dir}? [y|n] "))
            if(answer %in% c("y", "Y", "yes", "Yes", "YES")) {
                dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
            } else {
                stop_wrap(qq("not allowed to create @{output_dir}."))
            }
        } else {
            message("create output_dir: ", temp_dir)
            dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
        }
    }
    temp_dir = normalizePath(temp_dir)
    output_dir = normalizePath(output_dir)

    cat(cyan(qq("- job: '@{magenta(name)}' from script @{basename(script)}\n")))
    
    command = qq("@{bsub_opt$call_Rscript(R_version)} '@{script}' @{argv};")

    if(working_dir != "") {
        sh_head = paste0(sh_head, "\n", qq("cd '@{working_dir}'\n"))
    }

    bsub_submit(
        command = command, 
        hours = hours,
        memory = memory,
        cores = cores,
        name = name,
        name_uid = name_uid,
        output_dir = output_dir,
        temp_dir = temp_dir,
        dependency = dependency,
        enforce = enforce,
        local = local,
        sh_head = sh_head)
}


#' @param cmd A single-line command.
#' @param sh Path of the bash script.
#' @param env_var Environment variables. It should be a named vector. Note environment variables can also be directly set in `sh_head`.
#' @rdname bsub
#' @export
bsub_cmd = function(cmd, 
    sh = NULL,
    name = NULL, 
    hours = 1, 
    memory = 1, 
    cores = 1,
    temp_dir = bsub_opt$temp_dir,
    output_dir = bsub_opt$output_dir,
    dependency = NULL,
    enforce = bsub_opt$enforce, 
    local = bsub_opt$local,
    env_var = NULL,
    sh_head = bsub_opt$sh_head,
    ask = TRUE,
    ...) {

    if(!under_same_file_system()) {
        stop("Job can only be sumitted on the same file system as submission nodes.")
    }
    
    if(bsub_opt$ignore) return(invisible(NULL))

    if(!is.null(sh)) {
        if(!file.exists(sh)) {
            stop_wrap(qq("Cannot find the bash script: @{sh}"))
        }

        cmd = readLines(sh)
    }

    if(is.null(name)) {
        name = paste0("cmd_", digest::digest(cmd, "crc32"))
    } else {
        name = qq(name)
    }

    t = as.POSIXlt(Sys.time())
    t = as.numeric(t) + t$sec - floor(t$sec)
    t = gsub("\\.", "_", t)
    name_uid = paste0(name, "_", t)

    if(!file.exists(temp_dir)) {
        if(ask) {
            answer = readline(qq("create temp_dir: @{temp_dir}? [y|n] "))
            if(answer %in% c("y", "Y", "yes", "Yes", "YES")) {
                dir.create(temp_dir, recursive = TRUE, showWarnings = FALSE)
            } else {
                stop_wrap(qq("not allowed to create @{temp_dir}."))
            }
        } else {
            message("create temp_dir: ", temp_dir)
            dir.create(temp_dir, recursive = TRUE, showWarnings = FALSE)
        }
    }
    if(!file.exists(output_dir)) {
        if(ask) {
            answer = readline(qq("create output_dir: @{output_dir}? [y|n] "))
            if(answer %in% c("y", "Y", "yes", "Yes", "YES")) {
                dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
            } else {
                stop_wrap(qq("not allowed to create @{output_dir}."))
            }
        } else {
            message("create output_dir: ", temp_dir)
            dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)
        }
    }
    temp_dir = normalizePath(temp_dir)
    output_dir = normalizePath(output_dir)

    cat(cyan(qq("- job: '@{magenta(name)}' from a list of commands\n")))
    
    command = paste(cmd, collapse = "\n")

    if(!is.null(env_var)) {
        for(env_nm in names(env_var)) {
            sh_head = paste0(sh_head, "\n", env_nm, "=", env_var[env_nm])
        }
    }

    bsub_submit(
        command = command, 
        hours = hours,
        memory = memory,
        cores = cores,
        name = name,
        name_uid = name_uid,
        output_dir = output_dir,
        temp_dir = temp_dir,
        dependency = dependency,
        enforce = enforce,
        local = local,
        sh_head = sh_head)
}

bsub_submit = function(command,
    hours,
    memory,
    cores,
    name,
    name_uid,
    output_dir,
    temp_dir,
    dependency = NULL,
    enforce = TRUE,
    local = FALSE,
    sh_head = "") {

    if(!under_same_file_system()) {
        stop("Job can only be sumitted on submission nodes.")
    }

    output_dir = normalizePath(output_dir)
    temp_dir = normalizePath(temp_dir)

    done = qq("@{output_dir}/@{name}.done")
    old_flag = qq("@{output_dir}/@{name}.flag")
    pend = qq("@{output_dir}/@{name}.pend")
    run = qq("@{output_dir}/@{name}.run")
    output = qq("@{output_dir}/@{name_uid}.out")

    if(!enforce && (file.exists(done) || file.exists(old_flag))) {
        cat(cyan(qq("Job '@{name}' is already done, skip.\n")))
        return(invisible(NULL))
    }
    if(!enforce && file.exists(pend)) {
        # if the pending job is killed, the pend flag is still there
        if("PEND" %in% job_status_by_name(name)) {
            cat(cyan(qq("Job '@{name}' is pending, skip.\n")))
            return(invisible(NULL))
        }
    }
    if(!enforce && file.exists(run)) {
        # if the running job is killed, the run flag is still there
        if("RUN" %in% job_status_by_name(name)) {
            cat(cyan(qq("Job '@{name}' is running, skip.\n")))
            return(invisible(NULL))
        }
    }

    if(file.exists(done)) {
        file.remove(done);
    }

    
    sh_file = tempfile(paste0(name_uid, "_"), tmpdir = temp_dir, fileext = ".sh")
    con = file(sh_file, "w")

    writeLines("########## temporary bash script ###############", con)
    
    writeLines(qq("[ -f '@{pend}' ] && rm '@{pend}'\n"), con)  # remove the pend flag file
    writeLines(qq("touch '@{run}'\n"), con)  # add the running flag
    
    if(!identical(sh_head, "")) {
        writeLines("##### sh_head start #####", con)
        writeLines(sh_head, con)
        writeLines("##### sh_head end #####\n\n", con)
    }

    # wrap command with eval
    writeLines("##### commands start #####", con)
    writeLines(command, con)
    writeLines("##### commands end #####\n\n", con)

    # test the command, 
    writeLines(qq("
if [ $? -ne 0 ]
then
    rm '@{run}'
    echo Exit code is not equal to zero. There is an error.
    exit 666
fi"), con)

    writeLines(qq("rm '@{run}'\n"), con)
    writeLines(qq("touch '@{done}'\n"), con)
    close(con)

    if(local) {
        cmd = qq("bash @{sh_file}")
        cat(cmd, "\n")
        system(cmd)
        file.remove(sh_file)
        return(invisible(NULL))
    }

    system(qq("chmod 755 @{sh_file}"))

    cmd = bsub_opt$bsub_template(name, hours, memory, cores, output, bsub_opt$group)
    if(length(dependency)) {
        dependency_str = paste( paste("done(", dependency, ")"), collapse = " && " )
        cmd = qq("@{cmd} -w '@{dependency_str}' \\\n")
    }
    cmd = qq("@{cmd} < '@{sh_file}'")
    cat(silver(cmd), "\n")

    cmd = bsub_opt$bsub_template(name, hours, memory, cores, output, bsub_opt$group)
    # add values of name_uid and temp_dir to `bjobs` in job description
    cmd = qq("@{cmd} -env 'all,R_BSUB_NAME_UID=@{name_uid},R_BSUB_TEMP_DIR=@{temp_dir}' -Jd 'R_BSUB_NAME_UID=@{name_uid},R_BSUB_TEMP_DIR=@{temp_dir}'")
    if(length(dependency)) {
        dependency_str = paste( paste("done(", dependency, ")"), collapse = " && " )
        cmd = qq("@{cmd} -w '@{dependency_str}' \\\n")
    }
    cmd = qq("@{cmd} < '@{sh_file}'")

    file.create(pend)
    txt = run_cmd(cmd, print = FALSE)
    file.remove(sh_file)

    job_id = gsub("^.*<(\\d+)>.*$", "\\1", txt)

    return(job_id)
}


#' Submit a random job
#' 
#' @param name Job name.
#' @param secs Seconds to sleep.
#' @param ... Pass to [`bsub_chunk()`].
#' 
#' @details
#' It simply runs `Sys.sleep(secs)` in the job.
#' 
#' @export
#' @importFrom stats runif
#' @returns A job ID.
#' @examples
#' \dontrun{
#' random_job()
#' }
random_job = function(name, secs = 30, ...) {
    if(missing(name)) {
        name = paste0("R_random_job_", digest::digest(runif(1), "crc32"))
    }
    secs = secs
    bsub_chunk({
        Sys.sleep(secs)
    }, name = name, variables = "secs", ...)
}

#' Browser-based interactive job monitor 
#'
#' @details
#' The monitor is implemented as a shiny app.
#' @export
#' @examples
#' \dontrun{
#' monitor()
#' }
monitor = function() {

    pkgs = c("shiny", "DT", "shinyjs", "ggplot2", "igraph", "DiagrammeR")
    l = sapply(pkgs, requireNamespace, quietly = TRUE)

    if(any(!l)) {
        stop_wrap(qq("You need to install '@{paste(pkgs[!l], collapse = ', ')}' to use the monitor."))
    }
    
    if(!on_submission_node()) {
        ssh_validate()
    }

    suppressPackageStartupMessages(shiny::runApp(system.file("app", package = "bsub")))

    # if(identical(topenv(), asNamespace("bsub"))) {
    #     if(bsub_opt$verbose) cat("run job monitor from the package.\n")
    #     suppressPackageStartupMessages(shiny::runApp(system.file("app", package = "bsub")))
    # } else if(grepl("odcf", Sys.info()["nodename"])) {
    #     if(bsub_opt$verbose) cat("run job monitor from odcf node.\n")
    #     suppressPackageStartupMessages(shiny::runApp("/desktop-home/guz/project/development/bsub/inst/app"))
    # } else if(grepl("w610", Sys.info()["nodename"])) {
    #     if(bsub_opt$verbose) cat("run job monitor from w610 node.\n")
    #     suppressPackageStartupMessages(shiny::runApp("~/project/development/bsub/inst/app"))
    # } else {
    #     if(bsub_opt$verbose) cat("run job monitor from local laptop.\n")
    #     suppressPackageStartupMessages(shiny::runApp("~/project/development/bsub/inst/app"))
    # }
}

#' Visualize statistics of jobs
#' 
#' @param status Status of the jobs. Use "all" for all status.
#' @param filter Regular expression to filter on job names.
#' @param job_tb A data frame returned by [`bjobs()`]. Only internally used.
#'
#' @details
#' `bjobs_barplot()` draws barplots of number of jobs per day.
#' 
#' `bjobs_timeline()` draws segments of duration of jobs. In the plot, each segment represents
#' a job and the width of the segment correspond to its duration.
#'
#' @rdname visualize
#' @importFrom graphics text
#' @export
bjobs_barplot = function(status = c("RUN", "EXIT", "PEND", "DONE"), filter = NULL, job_tb = NULL) {
    if(is.null(job_tb)) {
        job_tb = bjobs(status = status, filter = filter, print = FALSE)
    } else {
        job_tb = job_tb[job_tb$STAT %in% status , , drop = FALSE]
    }
    job_tb$STAT = as.character(job_tb$STAT)
    job_tb$STAT[!job_tb$STAT %in% status] = "Others"
    job_tb$STAT = factor(job_tb$STAT, levels = intersect(c(status, "Others"), as.character(job_tb$STAT)))
    job_tb = job_tb[as.double(Sys.time() - job_tb$SUBMIT_TIME, units = "secs") <= 7*24*60*60, , drop = FALSE]

    if(nrow(job_tb) == 0) {
        plot(NULL, xlim = c(0, 1), ylim = c(0, 1), axes = FALSE, ann = FALSE)
        text(0.5, 0.5, "No job was found in the most recent week.")
        return(invisible(NULL))
    }

    suppressWarnings(p <- ggplot2::ggplot(job_tb, ggplot2::aes(x = as.Date(job_tb$SUBMIT_TIME), fill = job_tb$STAT)) + ggplot2::geom_bar(position=ggplot2::position_dodge()) +
        ggplot2::xlab("Submitted time") + ggplot2::ylab("Number of jobs") + ggplot2::labs(fill = "Status")
    )
    p = p + ggplot2::scale_fill_manual(breaks = names(STATUS_COL), values = STATUS_COL)
    tb = table(job_tb$STAT)
    tb =  tb[tb > 0]
    p = p + ggplot2::ggtitle(paste0(paste(qq("@{tb} @{names(tb)} job@{ifelse(tb == 1, '', 's')}", collapse = FALSE), collapse = ", "), " in the most recent one week"))
    print(p)
}

#' @rdname visualize
#' @importFrom graphics axis box legend par segments text title
#' @export
bjobs_timeline = function(status = c("RUN", "EXIT", "PEND", "DONE"), filter = NULL, job_tb = NULL) {
    if(is.null(job_tb)) {
        job_tb = bjobs(status = status, filter = filter, print = FALSE)
    } else {
        job_tb = job_tb[job_tb$STAT %in% status, , drop = FALSE]
    }
    job_tb$STAT = as.character(job_tb$STAT)
    job_tb$STAT[!job_tb$STAT %in% status] = "Others"
    job_tb$STAT = factor(job_tb$STAT, levels = intersect(c(status, "Others"), as.character(job_tb$STAT)))
    
    if(nrow(job_tb) == 0) {
        plot(NULL, xlim = c(0, 1), ylim = c(0, 1), axes = FALSE, ann = FALSE)
        text(0.5, 0.5, "No jobs found.")
        return(invisible(NULL))
    }

    units(job_tb$TIME_PASSED) = "secs"

    x1 = as.numeric(job_tb$START_TIME)
    x2 = x1 + as.numeric(job_tb$TIME_PASSED)
    now = as.numeric(Sys.time())
    x2[is.na(x2)] = now
    y = runif(nrow(job_tb))

    xlim = c(min(x1, na.rm = TRUE), max(x2, na.rm = TRUE))
    plot(NULL, xlim = xlim, ylim = c(0, 1), axes = FALSE, ann = FALSE)
    segments(x1, y, x2, y, col = STATUS_COL[as.character(job_tb$STAT)], lwd = 2)
    at = unique(as.Date(c(job_tb$START_TIME, job_tb$FINISH_TIME)))
    labels = as.character(at)
    at = as.numeric(as.POSIXlt(at))
    l = at >= xlim[1] & at <= xlim[2]
    at = at[l]
    labels = labels[l]
    at = c(at, now)
    labels = c(labels, "now")
    box()
    axis(side = 1, at = at, labels = labels)
    op = par("xpd")
    par(xpd = NA)
    col = STATUS_COL[intersect(names(STATUS_COL), as.character(job_tb$STAT))]
    legend(x = mean(par("usr")[1:2]), y = mean(par("usr")[4]), legend = names(col), 
        col = col, lty = 1, lwd = 2, ncol = length(col), xjust = 0.5, yjust = 0, cex = 0.6)

    tb = table(job_tb$STAT)
    tb =  tb[tb > 0]
    title(paste0(paste(qq("@{tb} @{names(tb)} job@{ifelse(tb == 1, '', 's')}", collapse = FALSE), collapse = ", "), " within one week"))

    par(xpd = op)
}

############
## mot used
############

# cmd1 = "echo 1 > @{test_file}"
# cmd2 = "echo 2 >> @{test_file}"
# cmd3 = "echo 3 >> @{test_file}"

# job1 = job_cmd(cmd1, name = "job1")
# job2 = job_cmd(cmd2, name = "job2")
# job3 = job_cmd(cmd3, name = "job3")

# jobs = list(job1 = job1, job2 = job2, job3 = job3)


# dep = pipeline_dep(
# 	"job1 -> job2",
#     "job2 -> job3"
# )

# test_file = "~/test.txt"
# pipeline(dep = dep, jobs = jobs)

job_cmd = function(cmd, name, sh_head = "", ...) {
	args = list(...)
	args$cmd = cmd
	args$name = name
	args$sh_head = ""
	args$ask = FALSE
	args
}

pipeline_dep = function(...) {
	dep = unlist(list(...))
	lt = strsplit(dep, "\\s*->?\\s*")

	dep = cbind(from = sapply(lt, function(x) x[1]),
		        to = sapply(lt, function(x) x[2]))

	all_jobs = unique(as.vector(dep))
	n_jobs = length(all_jobs)
	job2ind = structure(seq_along(all_jobs), names = all_jobs)

	lt_parents2 = split(unname(job2ind[dep[, 1]]), dep[, 2])
	lt_children2 = split(unname(job2ind[dep[, 2]]), dep[, 1])
	
	lt_parents = rep(list(integer(0)), n_jobs)
	lt_children = lt_parents
	lt_parents[ job2ind[names(lt_parents2)] ] = lt_parents2
	lt_children[ job2ind[names(lt_children2)] ] = lt_children2

	roots = which(vapply(lt_parents, length, FUN.VALUE = integer(1)) == 0)

	depth = dag_depth_bfs(lt_children, roots)

	list(all_jobs = all_jobs,
		 n_jobs = n_jobs,
		 lt_children = lt_children, 
		 lt_parents = lt_parents,
		 roots = roots,
		 depth = depth)

}

pipeline = function(dep, jobs, prefix = "", suffix = random_str(), 
	envir = parent.frame(), vars = NULL) {

	jobs_not_defined = setdiff(dep$all_jobs, names(jobs))
	if(length(jobs_not_defined)) {
		stop_wrap(qq("Job: @{paste(jobs_not_defined, collapse = ', ')} not defined."))
	}

	depth = dep$depth

	if(!is.null(vars)) {
		if(is.list(vars) || is.environment(vars)) {
			stop_wrap("If `vars` is set, it should be a list or an environment object.")
		}
	}

	job_submitted = rep(NA, dep$n_jobs)

	for(i in order(depth)) {  # tpl ordered
		if(is.na(job_submitted[i])) { 
			from = dep$lt_parents[[i]] 
			if(length(from)) {
				jobs[[i]]$dependency = job_submitted[from]
			}

			id = .run_job(jobs[[i]], prefix = prefix, suffix = suffix, envir = envir, vars = vars)
			job_submitted[[i]] = id
		}
	}
}

random_str = function() {
	paste0(sample(c(letters, LETTERS), 6), collapse = "")
}


dag_depth_bfs = function(lt_children, roots) {

	n_terms = length(lt_children)

	d = rep(0L, n_terms)
	current_nodes = roots
	while(length(current_nodes)) {
		for(cr in current_nodes) {
			children = lt_children[[cr]]
			if(length(children)) {
				d[children] = base::pmax(d[children], d[cr] + 1L)
			}
		}
		current_nodes = unique(unlist(lt_children[current_nodes]))
	}
	d
}

.run_job = function(args, prefix, suffix, envir, vars = NULL) {
	args$name = paste0(prefix, args$name, suffix)
	if(is.null(vars)) {
		args$cmd = GetoptLong::qq(args$cmd, envir = envir)
		args$sh_head = GetoptLong::qq(args$sh_head, envir = envir)
	} else {
		args$cmd = GetoptLong::qq(args$cmd, envir = as.environment(vars))
		args$sh_head = GetoptLong::qq(args$sh_head, envir = as.environment(vars))
	}

	do.call(bsub_cmd, args)
}

---
title: "Send R code/R scripts/shell commands to LSF cluster"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Send R code/R scripts/shell commands to LSF cluster}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


There are the following vignettes:

- [Send R code/R scripts/shell commands to LSF cluster](https://jokergoo.github.io/bsub/articles/bsub_intro.html)
- [Configure bsub package](https://jokergoo.github.io/bsub/articles/configure_bsub_package.html)
- [Establish two ssh connections to reach the submission node](https://jokergoo.github.io/bsub/articles/two_ssh.html)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monitor.R
\name{bjobs_barplot}
\alias{bjobs_barplot}
\alias{bjobs_timeline}
\title{Visualize statistics of jobs}
\usage{
bjobs_barplot(
  status = c("RUN", "EXIT", "PEND", "DONE"),
  filter = NULL,
  job_tb = NULL
)

bjobs_timeline(
  status = c("RUN", "EXIT", "PEND", "DONE"),
  filter = NULL,
  job_tb = NULL
)
}
\arguments{
\item{status}{Status of the jobs. Use "all" for all status.}

\item{filter}{Regular expression to filter on job names.}

\item{job_tb}{A data frame returned by \code{\link[=bjobs]{bjobs()}}. Only internally used.}
}
\description{
Visualize statistics of jobs
}
\details{
\code{bjobs_barplot()} draws barplots of number of jobs per day.

\code{bjobs_timeline()} draws segments of duration of jobs. In the plot, each segment represents
a job and the width of the segment correspond to its duration.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bjobs.R
\name{bjobs}
\alias{bjobs}
\alias{bjobs_raw}
\alias{brecent}
\alias{bjobs_running}
\alias{bjobs_pending}
\alias{bjobs_done}
\alias{bjobs_exit}
\title{Summary of jobs}
\usage{
bjobs(
  status = c("RUN", "PEND"),
  max = Inf,
  filter = NULL,
  print = TRUE,
  job_id = NULL
)

bjobs_raw(fields = "jobid stat job_name queue")

brecent(max = 20, filter = NULL)

bjobs_running(max = Inf, filter = NULL)

bjobs_pending(max = Inf, filter = NULL)

bjobs_done(max = Inf, filter = NULL)

bjobs_exit(max = Inf, filter = NULL)
}
\arguments{
\item{status}{Status of the jobs. Use "all" for all status}

\item{max}{Maximal number of recent jobs.}

\item{filter}{Regular expression on job names.}

\item{print}{Whether to print the table.}

\item{job_id}{A single job ID, internally used.}

\item{fields}{Supported output fields, check \url{https://www.ibm.com/docs/en/spectrum-lsf/10.1.0?topic=information-customize-job-output}.
The value can be a vector of field names or a single string that already includes output fields separated by space.}
}
\value{
A data frame with selected job summaries.
}
\description{
Summary of jobs
}
\details{
There is an additional column "RECENT" which is the order
for the job with the same name. 1 means the most recent job.

You can directly type \code{bjobs} without parentheses which runs \code{bjobs} with defaults.
\itemize{
\item \code{brecent} shows the most recent.
\item \code{bjobs_done} shows the "DONE" jobs.
\item \code{bjobs_exit} shows the "EXIT" jobs.
\item \code{bjobs_pending} shows the "PEND" jobs.
\item \code{bjobs_running} shows the "RUN" jobs.
}

\code{bjobs_raw()} returns the table from the original \verb{bsubs -a -o '...'} call.
}
\examples{
\dontrun{
bjobs # this is the same as bjobs()
bjobs() # all running and pending jobs
bjobs(status = "all") # all jobs
bjobs(status = "RUN") # all running jobs, you can also use `bjobs_running`
bjobs(status = "PEND") # all pending jobs, you can also use `bjobs_pending`
bjobs(status = "DONE") # all done jobs, you can also use `bjobs_done`
bjobs(status = "EXIT") # all exit jobs, you can also use `bjobs_exit`
bjobs(status = "all", max = 20) # last 20 jobs
bjobs(status = "DONE", filter = "example") # done jobs with name '.*example.*'
}
\dontrun{
brecent  # this is the same as `brecent()`
brecent() # last 20 jobs (from all status)
brecent(max = 50) # last 50 jobs
brecent(filter = "example") # last 20 jobs with name ".*example.*"
}
\dontrun{
bjobs_running  # this is the same as `bjobs_running()`
bjobs_running() # all running jobs
bjobs_running(max = 50) # last 50 running jobs
bjobs_running(filter = "example") # running jobs with name ".*example.*"
}
\dontrun{
bjobs_pending  # this is the same as `bjobs_pending()`
bjobs_pending() # all pending jobs
bjobs_pending(max = 50) # last 50 pending jobs
bjobs_pending(filter = "example") # pending jobs with name ".*example.*"
}
\dontrun{
bjobs_done  # this is the same as `bjobs_done()`
bjobs_done() # all done jobs
bjobs_done(max = 50) # last 50 done jobs
bjobs_done(filter = "example") # done jobs with name ".*example.*"
}
\dontrun{
bjobs_exit  # this is the same as `bjobs_exit()`
bjobs_exit() # all exit jobs
bjobs_exit(max = 50) # last 50 exit jobs
bjobs_exit(filter = "example") # exit jobs with name ".*example.*"
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status.R
\name{job_status_by_name}
\alias{job_status_by_name}
\alias{job_status_by_id}
\title{Job status by job ID or name}
\usage{
job_status_by_name(job_name)

job_status_by_id(job_id)
}
\arguments{
\item{job_name}{A single job name.}

\item{job_id}{A single job ID.}
}
\value{
A vector of job status, with job IDs are names.
}
\description{
Job status by job ID or name
}
\examples{
\dontrun{
job_status_by_name("example")
job_status_by_id(123456)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean.R
\name{list_dump_files}
\alias{list_dump_files}
\title{Check whether there are dump files}
\usage{
list_dump_files(print = TRUE)
}
\arguments{
\item{print}{Whether to print messages.}
}
\value{
A vector of file names.
}
\description{
Check whether there are dump files
}
\details{
For the failed jobs, LSF cluster might generate a core dump file and R might generate a \code{.RDataTmp} file.

Note if you manually set working directory in your R code/script, the R dump file can be not caught.
}
\examples{
\dontrun{
list_dump_files()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/job_log.R
\name{job_log}
\alias{job_log}
\title{Obtain Job log}
\usage{
job_log(job_id, print = TRUE, n_line = 10)
}
\arguments{
\item{job_id}{The job id. It can be a single job or a vector of job ids.}

\item{print}{Whether to print the log message to the terminal.}

\item{n_line}{Number of last lines for each job to show when multiple jobs are queried.}
}
\value{
The log messages as a vector.
}
\description{
Obtain Job log
}
\examples{
\dontrun{
# a single job
job_id = 1234567  # job ids can be get from `bjobs`
job_log(job_id)
# multiple jobs
job_id = c(10000000, 10000001, 10000002) 
job_log(job_id)  # by  default last 10 lines for each job are printed
job_log(job_id, n_line = 20) # print last 20 lines for each job
# logs for all running jobs
job_log()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bjobs.R
\name{bjobs_reset_timestamp}
\alias{bjobs_reset_timestamp}
\title{Clear job history}
\usage{
bjobs_reset_timestamp()
}
\description{
Clear job history
}
\details{
It sets a timestamp to only show jobs after it.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_cmd.R
\name{bkill}
\alias{bkill}
\title{Kill jobs}
\usage{
bkill(job_id, filter = NULL)
}
\arguments{
\item{job_id}{A vector of job IDs or a data frame returned by \code{\link[=bjobs]{bjobs()}}.}

\item{filter}{Regular expression on job names (only the running and pending jobs). It is
only used when \code{job_id} is not set.}
}
\description{
Kill jobs
}
\examples{
\dontrun{
job_id = c(10000000, 10000001, 10000002)  # job ids can be get from `bjobs()`
bkill(job_id)
# kill all jobs (running and pending) of which the names contain "example"
bkill(filter = "example") 
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bsub.R
\name{bsub_chunk}
\alias{bsub_chunk}
\alias{bsub_script}
\alias{bsub_cmd}
\title{Submit jobs}
\usage{
bsub_chunk(
  code,
  name = NULL,
  packages = bsub_opt$packages,
  image = bsub_opt$image,
  variables = character(),
  share = character(),
  working_dir = bsub_opt$working_dir,
  hours = 1,
  memory = 1,
  cores = 1,
  R_version = bsub_opt$R_version,
  temp_dir = bsub_opt$temp_dir,
  output_dir = bsub_opt$output_dir,
  dependency = NULL,
  enforce = bsub_opt$enforce,
  local = bsub_opt$local,
  script = NULL,
  start = NULL,
  end = NULL,
  save_var = FALSE,
  sh_head = bsub_opt$sh_head,
  ask = TRUE
)

bsub_script(
  script,
  argv = "",
  name = NULL,
  working_dir = bsub_opt$working_dir,
  hours = 1,
  memory = 1,
  cores = 1,
  R_version = bsub_opt$R_version,
  temp_dir = bsub_opt$temp_dir,
  output_dir = bsub_opt$output_dir,
  dependency = NULL,
  enforce = bsub_opt$enforce,
  local = bsub_opt$local,
  sh_head = bsub_opt$sh_head,
  ask = TRUE,
  ...
)

bsub_cmd(
  cmd,
  sh = NULL,
  name = NULL,
  hours = 1,
  memory = 1,
  cores = 1,
  temp_dir = bsub_opt$temp_dir,
  output_dir = bsub_opt$output_dir,
  dependency = NULL,
  enforce = bsub_opt$enforce,
  local = bsub_opt$local,
  env_var = NULL,
  sh_head = bsub_opt$sh_head,
  ask = TRUE,
  ...
)
}
\arguments{
\item{code}{The code chunk, it should be embraced by \code{{}}.}

\item{name}{If name is not specified, an internal name calculated by \code{\link[digest:digest]{digest::digest()}} on the chunk is automatically assigned.}

\item{packages}{A character vector with package names that will be loaded before running the script. There is a special name \verb{_in_session_}
that loads all the packages loaded in current R session.}

\item{image}{A character vector of \code{.RData}/\code{.rda} files that will be loaded before running the script.
When \code{image} is set to \code{TRUE}, all variables in \code{\link{.GlobalEnv}} will be saved
into a temporary file and all attached packages will be recorded. The temporary files
will be removed after the job is finished.}

\item{variables}{A character vector of variable names that will be loaded before running the script. There is a special name \verb{_all_functions_}
that saves all functions defined in the global environment.}

\item{share}{A character vector of variables names for which the variables are shared between jobs. Note the temporary \code{.RData} files are not deleted
automatically.}

\item{working_dir}{The working directory.}

\item{hours}{Running time of the job.}

\item{memory}{Memory usage of the job. It is measured in GB.}

\item{cores}{Number of cores.}

\item{R_version}{R version.}

\item{temp_dir}{Path of temporary folder where the temporary R/bash scripts will be put.}

\item{output_dir}{Path of output folder where the output/flag files will be put.}

\item{dependency}{A vector of job IDs that current job depends on.}

\item{enforce}{If a flag file for the job is found, whether to enforce to rerun the job.}

\item{local}{Run job locally (not submitting to the LSF cluster)?}

\item{script}{In \code{bsub_chunk()}, it is the path of a script where code chunks will be extracted and sent to the cluster.
It is always used with \code{start} and \code{end} arguments. In \code{bsub_script()}, it is the path of the R script to submit.}

\item{start}{A numeric vector that contains line indices of the starting code chunk or a character vector
that contain regular expression to match the start of code chunks.}

\item{end}{Same setting as \code{start}.}

\item{save_var}{Whether save the last variable in the code chunk? Later the variable
can be retrieved by \code{\link[=retrieve_var]{retrieve_var()}}.}

\item{sh_head}{Commands that are written as head of the sh script.}

\item{ask}{Whether to promote.}

\item{argv}{A string of command-line arguments.}

\item{...}{Command-line arguments can also be specified as name-value pairs.}

\item{cmd}{A single-line command.}

\item{sh}{Path of the bash script.}

\item{env_var}{Environment variables. It should be a named vector. Note environment variables can also be directly set in \code{sh_head}.}
}
\value{
A job ID.
}
\description{
Submit jobs
}
\details{
\code{job_chunk()} submits R code chunk.

\code{job_script()} submits R script with command-line arguments.

\code{job_cmd()} submits general bash commands.
}
\examples{
\dontrun{
bsub_chunk(name = "example", memory = 10, hours = 10, cores = 4,
{
    Sys.sleep(5)
})

# the R version is defined in bsub_opt$R_version
bsub_script("/path/of/foo.R", name = ..., memory = ..., cores = ..., ...)
# with command-line arguments
bsub_script("/path/of/foo.R", argv = "--a 1 --b 3", ...)

# put all arguments also in the command
bsub_cmd("some-tool -arg1 1 -arg2 2", name = ..., memory = ..., cores = ..., ...)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bsub.R
\name{random_job}
\alias{random_job}
\title{Submit a random job}
\usage{
random_job(name, secs = 30, ...)
}
\arguments{
\item{name}{Job name.}

\item{secs}{Seconds to sleep.}

\item{...}{Pass to \code{\link[=bsub_chunk]{bsub_chunk()}}.}
}
\value{
A job ID.
}
\description{
Submit a random job
}
\details{
It simply runs \code{Sys.sleep(secs)} in the job.
}
\examples{
\dontrun{
random_job()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global.R
\docType{data}
\name{bconf}
\alias{bconf}
\alias{print.bconf}
\title{Print current configurations}
\format{
An object of class \code{bconf} of length 1.
}
\usage{
bconf

\method{print}{bconf}(x, ...)
}
\arguments{
\item{x}{A \code{bconf} object.}

\item{...}{Other parameters.}
}
\value{
A \code{bconf} object.
}
\description{
Print current configurations
}
\details{
This function is only for printing. Use \code{\link[=bsub_opt]{bsub_opt()}} to change configurations.

You simply type \code{bconf} (without the brackets) in the interactive R console.
}
\examples{
bconf
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rerun.R
\name{job_rerun}
\alias{job_rerun}
\alias{pipeline_rerun}
\title{Rerun jobs}
\usage{
job_rerun(job_id, dependency = character(0), verbose = TRUE, job_tb = NULL)

pipeline_rerun(job_id, skip_done = TRUE, verbose = TRUE)
}
\arguments{
\item{job_id}{A single job ID. In \code{pipeline_rerun()}, it is the job ID of any one job in the pipeline.}

\item{dependency}{A vector of job IDs that current job depends on.}

\item{verbose}{Whether to print messages.}

\item{job_tb}{The data frame returned from \code{\link[=bjobs]{bjobs()}}, internally used.}

\item{skip_done}{Whether to skip done jobs.}
}
\value{
\code{job_rerun()} returns the job IDs. \code{pipeline_rerun()} returns \code{NULL}.
}
\description{
Rerun jobs
}
\details{
In \code{pipeline_rerun()}, the full set of jobs can be captured by one job in the pipeline.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bjobs.R
\name{print.bjobs}
\alias{print.bjobs}
\title{Summary of jobs}
\usage{
\method{print}{bjobs}(x, ...)
}
\arguments{
\item{x}{A \code{bjobs} object.}

\item{...}{other arguments.}
}
\description{
Summary of jobs
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ssh.R
\name{ssh_connect}
\alias{ssh_connect}
\alias{ssh_disconnect}
\title{Connect to submisstion via ssh}
\usage{
ssh_connect(...)

ssh_disconnect()
}
\arguments{
\item{...}{Pass to \code{\link[ssh:ssh]{ssh::ssh_connect()}}.}
}
\description{
Connect to submisstion via ssh
}
\examples{
# ssh is automatically connected. To manually connect ssh, run:
\dontrun{
ssh_connect()
ssh_disconnect()
}
# where the user name is the one you set in `bsub_opt$user` and
# the node is the one you set in `bsub_opt$login_node`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_cmd.R
\name{run_cmd}
\alias{run_cmd}
\title{Run command on submission node}
\usage{
run_cmd(cmd, print = FALSE)
}
\arguments{
\item{cmd}{A single-line command.}

\item{print}{Whether to print output from the command.}
}
\value{
The output of the command.
}
\description{
Run command on submission node
}
\details{
If current node is not the submission node, the command is executed via ssh.
}
\examples{
\dontrun{
# run pwd on remote node
run_cmd("pwd")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format.R
\name{formatFileSize}
\alias{formatFileSize}
\title{Format file size}
\usage{
formatFileSize(table, columns)
}
\arguments{
\item{table}{Internally used.}

\item{columns}{Internally used.}
}
\description{
Format file size
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retrieve.R
\name{retrieve_var}
\alias{retrieve_var}
\title{Retrieve saved variable}
\usage{
retrieve_var(job_id, job_name = NULL, wait = 30)
}
\arguments{
\item{job_id}{A single job ID.}

\item{job_name}{A single job name. Since jobs may have the same names, the most recent job is selected.}

\item{wait}{Seconds to wait until the job is finished.}
}
\value{
The retrieved object.
}
\description{
Retrieve saved variable
}
\details{
It retrieves the saved variable in \code{\link[=bsub_chunk]{bsub_chunk()}} when \code{save_rds = TRUE} is set.
}
\examples{
\dontrun{
job_id = bsub_chunk(name = "example", save_var = TRUE,
{
    Sys.sleep(10)
    1+1
})
retrieve_var(job_id)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dependency.R
\name{job_dependency_all}
\alias{job_dependency_all}
\alias{job_dependency_igraph}
\alias{job_dependency_dot}
\alias{job_dependency_diagram}
\title{Job dependencies}
\usage{
job_dependency_all(job_tb = NULL)

job_dependency_igraph(job_id, job_tb = NULL)

job_dependency_dot(job_id, job_tb = NULL, use_label = FALSE, label_width = 15)

job_dependency_diagram(
  job_id,
  job_tb = NULL,
  use_label = FALSE,
  label_width = 15,
  ...
)
}
\arguments{
\item{job_tb}{A data frame from \code{\link[=bjobs]{bjobs()}}. Internally used.}

\item{job_id}{A single job ID.}

\item{use_label}{Whether to use job names on the diagram?}

\item{label_width}{Max number of characters for wrapping the label into lines.}

\item{...}{Pass to \code{\link[DiagrammeR:grViz]{DiagrammeR::grViz()}}, such as the size of the html widget.}
}
\value{
\code{job_dependency_all()} returns a list that contains three elements:
\itemize{
\item \code{dep_mat}: a two column matrix containing dependencies from parents to children.
\item \code{id2name}: a named vector containing mapping from job IDs to job names.
\item \code{id2stat}: a named vector containing mapping from job IDs to job status.
}

\code{job_dependency_igraph()} returns a \code{\link[igraph:aaa-igraph-package]{igraph::igraph}} object which contains a dependency
graph induced by the input job ID.

\code{job_dependency_dot()} returns a DOT code for GraphViz visualization.

\code{job_dependency_diagram()} makes a HTML-based dependency diagram.
}
\description{
Job dependencies
}
\examples{
\dontrun{
job1 = random_job()
job2 = random_job()
job3 = random_job(dependency = c(job1, job2))

job_dependency_all()
job_dependency_igraph(job3)
cat(job_dependency_dot(job3))
job_dependency_diagram(job3)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean.R
\name{list_temp_files}
\alias{list_temp_files}
\alias{remove_temp_files}
\title{Clear temporary dir}
\usage{
list_temp_files()

remove_temp_files(ask = TRUE)
}
\arguments{
\item{ask}{Whether to promote.}
}
\description{
Clear temporary dir
}
\details{
The temporary files might be used by the running/pending jobs. Deleting them would affect some jobs.
You should better delete them after all jobs are done.
}
\examples{
\dontrun{
list_temp_files()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/monitor.R
\name{monitor}
\alias{monitor}
\title{Browser-based interactive job monitor}
\usage{
monitor()
}
\description{
Browser-based interactive job monitor
}
\details{
The monitor is implemented as a shiny app.
}
\examples{
\dontrun{
monitor()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/global.R
\name{bsub_opt}
\alias{bsub_opt}
\title{Parameters for bsub}
\usage{
bsub_opt(..., RESET = FALSE, READ.ONLY = NULL, LOCAL = FALSE, ADD = FALSE)
}
\arguments{
\item{...}{Arguments for the parameters, see "details" section}

\item{RESET}{reset to default values}

\item{READ.ONLY}{please ignore}

\item{LOCAL}{please ignore}

\item{ADD}{please ignore}
}
\value{
The corresponding option values.
}
\description{
Parameters for bsub
}
\details{
There are following global parameters:
\itemize{
\item \code{packages}: A character vector with package names that will be loaded before running the script.
\item \code{image}: A character vector of RData/rda files that will be loaded before running the script.
\item \code{temp_dir}: Path of temporary folder where the temporary R/bash scripts will be put.
\item \code{output_dir}: Path of output folder where the output/flag files will be put.
\item \code{enforce}: If a flag file for the job is found, whether to enforce to rerun the job.
\item \code{R_version}: The version of R.
\item \code{working_dir}: The working directory.
\item \code{ignore}: Whether ignore \code{bsub_chunk}, \code{bsub_script} and \code{bsub_cmd}.
\item \code{local}: Run job locally (not submitting to the LSF cluster)?
\item \code{call_Rscript}: How to call \code{Rscript} by specifying an R version number.
\item \code{submission_node}: A list of node names for submitting jobs.
\item \code{login_node}: This value basically is the same as \code{submission_node} unless the login nodes are different from submission nodes.
\item \code{sh_head}: Commands that are written as head of the sh script.
\item \code{user}: Username on the submission node.
\item \code{group}: The user group
\item \code{ssh_envir}: The commands for setting bash environment for successfully running bjobs, bsub, ...
\item \code{bsub_template}: Template for constructing \code{bsub} command.
\item \code{parse_time}: A function that parses time string from the LSF \code{bjobs} command to a \code{POSIXct} object.
\item \code{verbose}: Whether to print more messages.
}

\code{ssh_envir} should be properly set so that LSF binaries such as \code{bsub} or \code{bjobs} can be properly found.
There are some environment variables initialized when logging in the bash terminal while they are not initialized with the
ssh connection. Thus, some environment variables should be manually set.

An example for \code{ssh_envir} is as follows. The \code{LSF_ENVDIR} and \code{LSF_SERVERDIR} should be defined and exported.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{c("source /etc/profile",
  "export LSF_ENVDIR=/opt/lsf/conf",
  "export LSF_SERVERDIR=/opt/lsf/10.1/linux3.10-glibc2.17-x86_64/etc")
}\if{html}{\out{</div>}}

The values of these two variables can be obtained by entering following commands in your bash terminal (on the submission node):

\if{html}{\out{<div class="sourceCode">}}\preformatted{echo $LSF_ENVDIR
echo $LSF_SERVERDIR
}\if{html}{\out{</div>}}

The time strings by LSF \code{bjobs} command might be different for different configurations. The \strong{bsub}
package needs to convert the time strings to \code{POSIXlt} objects for calculating the time difference. Thus, if
the default time string parsing fails, users need to provide a user-defined function and set with \code{parse_time}
option in \code{bsub_opt}. The function accepts a vector of time strings and returns a \code{POSIXlt} object. For example,
if the time string returned from \code{bjobs} command is in a form of \verb{Dec 1 18:00:00 2019}, the parsing function
can be defined as:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{bsub_opt$parse_time = function(x) \{
    as.POSIXlt(x, format = "\\\\\%b \\\\\%d \\\\\%H:\\\\\%M:\\\\\%S \\\\\%Y")
\}
}\if{html}{\out{</div>}}
}
\examples{
# The default bsub_opt
bsub_opt
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{config_bsub}
\alias{config_bsub}
\title{Configure bsub global options}
\usage{
config_bsub(verbose = TRUE)
}
\arguments{
\item{verbose}{Whether to print messages.}
}
\description{
Configure bsub global options
}
\details{
It sets the submission nodes, user name and how to call \code{Rscript}.
}
